; A 6502 MMU emulator
; /Mic, 2004


.486
.model flat,stdcall
option casemap:none


include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\user32.inc

includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\user32.lib
wsprintfA PROTO C :DWORD, :VARARG
wsprintf equ <wsprintfA>

.data

	CPU_IRQ_CALLBACK		equ 000h
	CPU_SET_IO_CALLBACK		equ 001h
	CPU_IO_HANDLER			equ 001h
	CPU_CYCLE_COUNT			equ 002h
	CPU_REG_PC			equ 003h
	CPU_REG_SP			equ 004h
	CPU_BREAKPOINT			equ 005h
	CPU_SPEED			equ 006h
	CPU_INPB_CALLBACK		equ 007h
	CPU_OUTPB_CALLBACK		equ 008h
	CPU_CYCLE_SCALING		equ 009h
	
	MMU_READ_BYTE_CALLBACK		equ 100h
	MMU_READ_WORD_CALLBACK		equ 101h
	MMU_WRITE_BYTE_CALLBACK		equ 102h
	MMU_WRITE_WORD_CALLBACK		equ 103h
	MMU_TRANSLATE_ADDRESS_CALLBACK	equ 104h
	MMU_MAP_PORT			equ 105h
	
	GPU_CURRENT_SCANLINE		equ 200h
	GPU_SCREEN_WIDTH		equ 201h
	GPU_SCREEN_HEIGHT		equ 202h
	GPU_FRAME_COMPLETED		equ 203h
	GPU_SCREEN_PITCH		equ 204h
	GPU_MEMORY			equ 206h
	GPU_NUMBER_COLORS		equ 207h
	GPU_PALETTE			equ 208h
	GPU_VBLANK_ACTION		equ 209h
	GPU_HBLANK_ACTION		equ 20Ah

	ZOMBIE_WINDOW_HANDLE		equ 500h
	ZOMBIE_ROM_BASE			equ 501h
	
	PLUGIN_NAME			equ 600h
	
	INPUT		equ 0
	OUTPUT		equ 1

	ASK		equ 0
	TELL		equ 1

	JUMP		equ 0
	VECTOR		equ 0
	

	RAMOFFSET	equ 0000h
	RAMSIZE		equ 8000h
	
	ROMOFFSET	equ 8000h
	PAGESIZE	equ 4000h
	NUMPAGES	equ 2
	PAGESHIFT	equ 14
	PAGEREGBASE	equ 0FFF8h
	

	VRAMOFFSET	equ 6000000h
	VRAMSIZE	equ 40000h
	

	

	; Various prototypes
	_1ARGS_PROTO	typedef PROTO :DWORD
	_2ARGS_PROTO	typedef PROTO :DWORD,:DWORD
	_4ARGS_PROTO	typedef PROTO :DWORD,:DWORD,:DWORD,:DWORD
	_1ARGS		typedef ptr _1ARGS_PROTO
	_2ARGS		typedef ptr _2ARGS_PROTO
	_4ARGS		typedef ptr _4ARGS_PROTO


	;mmu_translate_address	PROTO :DWORD
	;mmu_read_byte	PROTO :DWORD
	;mmu_read_word	PROTO :DWORD
	;mmu_write_byte	PROTO :DWORD,:DWORD
	mmu_write_word	PROTO :DWORD,:DWORD
	
	outpb_30	PROTO :DWORD
	outpb_31	PROTO :DWORD
	outpb_32	PROTO :DWORD


	mmuName db "6502 MMU",0


	szBuffer	db 80 DUP(0)
	szFormat db "Setting io handler for port %02x, mode %d",13,10,0
	hout dd 0
	zero dd 0

.data?
	hInst		dd ?
	RAM		dd ?
	VRAM		dd ?
	cart		dd ?
	addressShift	dd ?
	addressLatch	dd ?
	dmaSrcAddress	dd ?
	dmaDstAddress	dd ?
	
	dummy 		dd ?
	pageTable 	dd 2 dup(?)
	iomap 		db 65536 dup(?)

	zombie		_4ARGS ?
	cpu_inpb 	_1ARGS ?
	cpu_outpb 	_2ARGS ?
	
	
.code



;-------------------------------------------------------------------------------------------------

	
LibMain PROC hInstDLL:DWORD, reason:DWORD, unused:DWORD

        .if reason == 1		;DLL_PROCESS_ATTACH
        	mov eax,hInstDLL
        	mov hInst,eax
		mov eax,TRUE
		ret

        .elseif reason == 0	;DLL_PROCESS_DETACH
        .elseif reason == 2	;DLL_THREAD_ATTACH
        .elseif reason == 3	;DLL_THREAD_DETACH
        .endif

        mov eax,FALSE
        ret
LibMain ENDP





mmu_init PROC commline:DWORD
	pusha
	
	mov eax,commline
	mov zombie,eax
	
	invoke GetStdHandle,STD_OUTPUT_HANDLE
	mov hout,eax

	invoke LocalAlloc,LPTR,RAMSIZE
	mov RAM,eax
	
	mov eax,0FFFFFFFFh
	mov ecx,4000h
	mov edi,OFFSET iomap
	rep stosd

	popa
	ret
mmu_init ENDP



mmu_reset PROC
	pusha

	invoke zombie,ASK,CPU_INPB_CALLBACK,0,0
	mov cpu_inpb,eax
	invoke zombie,ASK,CPU_OUTPB_CALLBACK,0,0
	mov cpu_outpb,eax

	invoke zombie,ASK,ZOMBIE_ROM_BASE,0,0
	mov cart,eax
	
	invoke zombie,TELL,GPU_VBLANK_ACTION,VECTOR,0FFFAh
	invoke zombie,TELL,GPU_HBLANK_ACTION,VECTOR,0FFFAh
	
	invoke zombie,ASK,GPU_MEMORY,0,0
	mov VRAM,eax

	invoke zombie,TELL,CPU_IO_HANDLER,130h,OFFSET outpb_30
	invoke zombie,TELL,CPU_IO_HANDLER,131h,OFFSET outpb_31
	invoke zombie,TELL,CPU_IO_HANDLER,132h,OFFSET outpb_32

	; Map ports
	mov byte ptr [iomap+0FF30h],30h
	mov byte ptr [iomap+0FF31h],31h
	mov byte ptr [iomap+0FF32h],32h
	
	mov dmaSrcAddress,0
	mov dmaDstAddress,0
	mov addressLatch,0
	mov addressShift,0

	mov edi,RAM
	mov al,0
	mov ecx,RAMSIZE
	rep stosb
	
	mov eax,cart
	mov ecx,0
	init_pagetable:
		mov [pageTable+ecx*4],eax
		add eax,PAGESIZE
		inc ecx
		cmp ecx,NUMPAGES
		jne init_pagetable
		
	popa
	ret
mmu_reset ENDP


mmu_close PROC
	invoke LocalFree,RAM
	ret
mmu_close ENDP


mmu_config PROC hWnd:DWORD
	ret
mmu_config ENDP



mmu_get PROC what:DWORD
	.if what == MMU_READ_BYTE_CALLBACK
		mov eax,offset mmu_read_byte
	.elseif what == MMU_READ_WORD_CALLBACK
		mov eax,offset mmu_read_word
	.elseif what == MMU_WRITE_BYTE_CALLBACK
		mov eax,offset mmu_write_byte
	.elseif what == MMU_WRITE_WORD_CALLBACK
		mov eax,offset mmu_write_word
	.elseif what == MMU_TRANSLATE_ADDRESS_CALLBACK
		mov eax,offset mmu_translate_address
	.elseif what == PLUGIN_NAME
		mov eax,offset mmuName
	.endif
	ret
mmu_get ENDP



mmu_set PROC what:DWORD,param1:DWORD,param2:DWORD
	.if what == MMU_MAP_PORT
		push ebx
		mov eax,param1
		mov ebx,param2
		mov [iomap+ebx],al
		pop ebx
	.endif
	ret
mmu_set ENDP


PUBLIC mmu_translate_address
mmu_translate_address: 			; address:DWORD
	mov eax,[esp+4]			; address
	.if eax >= ROMOFFSET
		push ebx
		sub eax,ROMOFFSET
		mov ebx,eax
		shr eax,PAGESHIFT
		mov eax,[pageTable+eax*4]
		and ebx,PAGESIZE-1
		add eax,ebx
		pop ebx
		ret 4
	.endif
	
	and eax,RAMSIZE-1
	add eax,RAM
	ret 4


PUBLIC mmu_read_byte
mmu_read_byte: 				; address:DWORD
	mov eax,[esp+4]			; address
	push ebx
	

	movzx ebx,byte ptr [iomap+eax]
	.if ebx != 0FFh
		invoke cpu_inpb,ebx
		pop ebx
		ret 4
	.endif

	.if eax >= ROMOFFSET
		sub eax,ROMOFFSET
		mov ebx,eax
		shr eax,PAGESHIFT
		mov eax,[pageTable+eax*4]
		and ebx,PAGESIZE-1
		add ebx,eax 
		movzx eax,byte ptr [ebx]
		pop ebx
		ret 4
	.endif
	
	and eax,RAMSIZE-1
	add eax,RAM
	movzx eax,byte ptr [eax]
	

	pop ebx

	ret 4


PUBLIC mmu_read_word
mmu_read_word: 				; address:DWORD
	push edx
	push ebx
	mov edx,[esp+12]		; address
	
	.if edx == 0FFh
		push edx
		call mmu_read_byte
		mov edx,0
		mov ebx,eax
		push edx
		call mmu_read_byte
		shl eax,8
		or eax,ebx
	.else	
		push edx
		call mmu_read_byte
		inc edx
		mov ebx,eax
		push edx
		call mmu_read_byte
		shl eax,8
		or eax,ebx
	.endif

	pop ebx
	pop edx
	ret 4


PUBLIC mmu_write_byte
mmu_write_byte: 			; address:DWORD,b:DWORD
	push ebx
	

	mov eax,[esp+8] 		; address
	movzx ebx,byte ptr [iomap+eax]
	.if ebx != 0FFh
		push dword ptr [esp+12]	; b
		push ebx		; port#
		call cpu_outpb
		mov eax,[esp+8]		; reload address
	.endif
	
	.if eax<ROMOFFSET
		mov ebx,eax
		mov eax,[esp+12]	; b
		and ebx,RAMSIZE-1
		add ebx,RAM
		mov [ebx],al
	.else
		.if eax>=PAGEREGBASE
			.if eax<PAGEREGBASE+NUMPAGES
				mov ebx,eax
				sub ebx,PAGEREGBASE
				mov eax,[esp+12]
				and eax,NUMPAGES-1
				shl eax,PAGESHIFT
				add eax,cart
				mov [pageTable+ebx*4],eax
			.endif
		.endif
	.endif
				
	pop ebx
	ret 8


mmu_write_word PROC address:DWORD,w:DWORD
	push ebx
	
	.if address == 0FFh
		mov eax,w
		mov ebx,eax
		and eax,0FFh
		push eax
		push address
		call mmu_write_byte
		shr ebx,8
		push ebx
		push zero
		call mmu_write_byte
	.else	
		mov eax,w
		mov ebx,eax
		and eax,0FFh
		push eax
		push address
		call mmu_write_byte
		inc dword ptr address
		shr ebx,8
		push ebx
		push address
		call mmu_write_byte
	.endif

	pop ebx
	ret
mmu_write_word ENDP



; DMA source address port
outpb_30 PROC b:DWORD
	push ecx
	
	mov eax,b
	mov ecx,addressShift
	and eax,0FFh
	shl eax,cl
	or addressLatch,eax

	add addressShift,8
	.if addressShift == 32
		mov eax,addressLatch
		mov addressShift,0
		mov addressLatch,0
		mov dmaSrcAddress,eax
	.endif
	
	pop ecx
	ret
outpb_30 ENDP



; DMA destination address port
outpb_31 PROC b:DWORD
	push ecx
	
	mov eax,b
	mov ecx,addressShift
	and eax,0FFh
	shl eax,cl
	or addressLatch,eax

	add addressShift,8
	.if addressShift == 32
		mov eax,addressLatch
		mov addressShift,0
		mov addressLatch,0
		mov dmaDstAddress,eax
	.endif
	
	pop ecx
	ret
outpb_31 ENDP



; DMA size port
outpb_32 PROC b:DWORD
	pusha
	
	mov eax,dmaSrcAddress
	.if eax >= ROMOFFSET
		.if eax < ROMOFFSET+PAGESIZE*NUMPAGES
			sub eax,ROMOFFSET
			mov ebx,eax
			shr ebx,PAGESHIFT
			mov esi,[pageTable+ebx*4]
			and eax,PAGESIZE-1
			add esi,eax
			jmp @@got_src
		.endif
	.endif
	popa
	ret
	
	@@got_src:

	mov eax,dmaDstAddress
	.if eax >= VRAMOFFSET
		.if eax < VRAMOFFSET+VRAMSIZE
			sub eax,VRAMOFFSET
			mov edi,VRAM
			add edi,eax
			jmp @@got_dst
		.endif
	.endif
	popa

	ret
	
	@@got_dst:

	mov ecx,b
	and ecx,0FFh
	shl ecx,6
	rep movsd
	
	popa
	ret
outpb_32 ENDP


END LibMain
