// ARM7 MMU emulator
// Mic, 2004

#include <windows.h>
#include <commctrl.h>
#include <string.h>
#include "zombie.h"


#define JUMP 0

HINSTANCE hInst;
HANDLE hout;
unsigned char *ROM,*IWRAM,*EXRAM,*cart,*VRAM;
unsigned char *iomap,**mmap;
int romOffset,romSize;
int iwramOffset,iwramSize;
int exramOffset,exramSize;
int vramOffset,vramSize;
char mmuName[] = "ARM7 MMU";
int dummy;
cpu_inpb_callback cpu_inpb;
cpu_outpb_callback cpu_outpb;
zombie_callback zombie;
char buffer[80];


BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);



bool __stdcall DllMain(HINSTANCE hDLL,DWORD reason,LPVOID unused)
{
	if (reason==DLL_PROCESS_ATTACH)
		hInst = hDLL;

	return true;
}



bool __stdcall mmu_init(zombie_callback commlink)
{
	zombie = commlink;

	romOffset = 0x8000000;
	romSize = 0x8000000;

	exramOffset = 0x2000000; 
	exramSize = 0x40000;

	iwramOffset = 0x3000000; 
	iwramSize = 0x8000;

	vramOffset = 0x6000000;
	vramSize = 0x40000;

	IWRAM = (unsigned char*)LocalAlloc(LPTR,iwramSize);
	EXRAM = (unsigned char*)LocalAlloc(LPTR,exramSize);

	iomap = new unsigned char[0x10000];

	for (int i=0; i<0x10000; i++)
		iomap[i] = 0xFF;

	hout = GetStdHandle(STD_OUTPUT_HANDLE);
	return true;
}


bool __stdcall mmu_reset()
{
	cpu_inpb = (cpu_inpb_callback)zombie(ASK,CPU_INPB_CALLBACK,0,0);
	cpu_outpb = (cpu_outpb_callback)zombie(ASK,CPU_OUTPB_CALLBACK,0,0);

	cart = (unsigned char*)zombie(ASK,ZOMBIE_ROM_BASE,0,0);
	ROM = cart;

	zombie(TELL,GPU_VBLANK_ACTION,(void*)JUMP,(void*)0x38);
	zombie(TELL,GPU_HBLANK_ACTION,(void*)JUMP,(void*)0x38);

	zombie(TELL,CPU_REG_PC,(void*)0x8000000,0);
	VRAM = (unsigned char*)zombie(ASK,GPU_MEMORY,0,0);

	memset(IWRAM,0,iwramSize);

	return true;
}


void __stdcall mmu_close()
{
	LocalFree(IWRAM);
	LocalFree(EXRAM);
}



int __stdcall mmu_config(HWND hwnd)
{
	return 0; /*DialogBoxParam(hInst,
            MAKEINTRESOURCE(IDD_MMU_CONFIG), hwnd,
            DlgProc, NULL);*/
}




unsigned char* __stdcall mmu_translate_address(int address)
{
	switch (address>>24)
	{
		case 8:
			return &ROM[address-romOffset];
		case 3:
			return &IWRAM[address-iwramOffset];
		case 2:
			return &EXRAM[address-exramOffset];
		case 6:
			return &VRAM[address-vramOffset];
	}

	return 0;
}



unsigned char __stdcall mmu_read_byte(int address)
{
	if (address < 0x10000)
		if (iomap[address] != 0xFF)
			return cpu_inpb(iomap[address]);

	switch (address>>24)
	{
		case 3:
			return IWRAM[address-iwramOffset];
		case 2:
			return EXRAM[address-exramOffset];
		case 6:
			return VRAM[address-vramOffset];
		case 8:
			return ROM[address-romOffset];
	}

	return 0;
}


unsigned short int __stdcall mmu_read_word(int address)
{
	unsigned short int *pS;

	switch (address>>24)
	{
		case 3:
			pS = (unsigned short int*)&IWRAM[address-iwramOffset];
			break;
		case 2:
			pS = (unsigned short int*)&EXRAM[address-exramOffset];
			break;
		case 6:
			pS = (unsigned short int*)&VRAM[address-vramOffset];
			break;
		case 8:
			pS = (unsigned short int*)&ROM[address-romOffset];
			break;
	}

	return *(pS);
}


int __stdcall mmu_read_dword(int address)
{
	int *pI;

	switch (address>>24)
	{
		case 8:
			pI = (int*)&ROM[address-romOffset];
			break;
		case 3:
			pI = (int*)&IWRAM[address-iwramOffset];
			break;
		case 2:
			pI = (int*)&EXRAM[address-exramOffset];
			break;
		case 6:
			pI = (int*)&VRAM[address-vramOffset];
			break;
	}


	return *(pI);
}


void __stdcall mmu_write_byte(int address,int byte)
{
	if (address < 0x10000)
		if (iomap[address] != 0xFF)
			cpu_outpb(iomap[address],byte);

	switch (address>>24)
	{
		case 3:
			IWRAM[address-iwramOffset] = (unsigned char)byte;
			break;
		case 2:
			EXRAM[address-exramOffset] = (unsigned char)byte;
			break;
		case 6:
			VRAM[address-vramOffset] = (unsigned char)byte;
			break;
	}

}


void __stdcall mmu_write_word(int address,int word)
{
	/*wsprintf(buffer,"mmu_write_word(%08x,%04x)\n",address,word);
	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/

	mmu_write_byte(address,word&0xFF);
	mmu_write_byte(address+1,word>>8);
}


void __stdcall mmu_write_dword(int address,int dwrd)
{
	int *pI;

	if (address==0x4000000)
	{
		wsprintf(buffer,"Writing %08x to %08x\n",dwrd,address);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
		return;
	}

	switch (address>>24)
	{
		case 3:
			pI = (int*)&IWRAM[address-iwramOffset];
			break;
		case 2:
			pI = (int*)&EXRAM[address-exramOffset];
			break;
		case 6:
			pI = (int*)&VRAM[address-vramOffset];
			break;
		default:
			return;
	}


	*(pI) = dwrd;
}


int __stdcall mmu_get(int what)
{
	switch (what)
	{
		case MMU_READ_BYTE_CALLBACK:
			return (int)mmu_read_byte;
		case MMU_READ_WORD_CALLBACK:
			return (int)mmu_read_word;
		case MMU_READ_DWORD_CALLBACK:
			return (int)mmu_read_dword;
		case MMU_WRITE_BYTE_CALLBACK:
			return (int)mmu_write_byte;
		case MMU_WRITE_WORD_CALLBACK:
			return (int)mmu_write_word;
		case MMU_WRITE_DWORD_CALLBACK:
			return (int)mmu_write_dword;
		case MMU_TRANSLATE_ADDRESS_CALLBACK:
			return (int)mmu_translate_address;
		case PLUGIN_NAME:
			return (int)&mmuName;
		default:
			break;
	}
	return 0;
}


void __stdcall mmu_set(int what,void* param1,void* param2)
{
	static char buffer[64];

	switch (what)
	{
		case MMU_MAP_PORT:
			iomap[(int)param2] = (unsigned char)param1;

		default:
			break;
	}
}


