; An ARM7TDMI emulator
; /Mic, 2004


.586p
.model flat,stdcall
option casemap:none


include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\user32.inc

includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\user32.lib

wsprintfA PROTO C :DWORD, :VARARG
wsprintf equ <wsprintfA>

.data

	CPU_IRQ_CALLBACK		equ 000h
	CPU_SET_IO_CALLBACK		equ 001h
	CPU_IO_HANDLER			equ 001h
	CPU_CYCLE_COUNT			equ 002h
	CPU_REG_PC			equ 003h
	CPU_REG_SP			equ 004h
	CPU_BREAKPOINT			equ 005h
	CPU_SPEED			equ 006h
	CPU_INPB_CALLBACK		equ 007h
	CPU_OUTPB_CALLBACK		equ 008h
	CPU_CYCLE_SCALING		equ 009h
	CPU_REG				equ 00Ah
	
	MMU_READ_BYTE_CALLBACK		equ 100h
	MMU_READ_WORD_CALLBACK		equ 101h
	MMU_WRITE_BYTE_CALLBACK		equ 102h
	MMU_WRITE_WORD_CALLBACK		equ 103h
	MMU_TRANSLATE_ADDRESS_CALLBACK	equ 104h
	MMU_READ_DWORD_CALLBACK		equ 106h
	MMU_WRITE_DWORD_CALLBACK	equ 107h

	GPU_CURRENT_SCANLINE		equ 200h
	GPU_SCREEN_WIDTH		equ 201h
	GPU_SCREEN_HEIGHT		equ 202h
	GPU_FRAME_COMPLETED		equ 203h
	GPU_SCREEN_PITCH		equ 204h

	ZOMBIE_WINDOW_HANDLE		equ 500h

	PLUGIN_NAME			equ 600h
	
	INPUT		equ 0
	OUTPUT		equ 1

	ASK		equ 0
	TELL		equ 1


	_1ARGS_PROTO	typedef PROTO :DWORD
	_2ARGS_PROTO	typedef PROTO :DWORD,:DWORD
	_4ARGS_PROTO	typedef PROTO :DWORD,:DWORD,:DWORD,:DWORD
	_1ARGS		typedef ptr _1ARGS_PROTO
	_2ARGS		typedef ptr _2ARGS_PROTO
	_4ARGS		typedef ptr _4ARGS_PROTO

	cpu_inpb	PROTO :DWORD
	cpu_outpb	PROTO :DWORD,:DWORD

	
	Overflow	db 0
	Carry		db 0
	Zero		db 0
	Negative	db 0

	flags		db 0,0,0,0,0,0,0,1
	cycle		dd 0
	breakpoint	dd 0
	speed		dd 0C00000h	; 12 MHz
	cpuName		db "ARM7TDMI",0

	szBranch db "Branch instruction",13,10,0
	szTrue db "true",13,10,0
	szBuffer db 64 dup (0)
	szFormat db "Read %08x at pc=%08x",13,10,0
	szFormat2 db "Instruction type %d at %08x",13,10,0
	szFormat3 db "Writing %04x at %08x",13,10,0
	szFormat4 db "Branching with link from %08x to %08x",13,10,0
	szFormat5 db "mov pc,lr (%08x)",13,10,0
	szFormat6 db "stmfd %08x",13,10,0
	szFormat7 db "ldmfd %08x",13,10,0
	szFormat8 db "ldr r%d,=%d",13,10,0
	szFormat9 db "MUL r%d,%08x,%08x",13,10,0
	
	
.data?
	r0	dd ?
	r1	dd ?
	r2	dd ?
	r3	dd ?
	r4	dd ?
	r5	dd ?
	r6	dd ?
	r7	dd ?
	r8	dd ?
	r9	dd ?
	r10	dd ?
	r11	dd ?
	r12	dd ?
	r13	dd ?	; SP
	r14	dd ?	; LR
	r15	dd ?	; PC
	r16	dd ?	; CPSR
	r17	dd ?
	
	hInst	dd ?
	hout	dd ?
	dummy 	dd ?
	lastPC 	dd ?
	physaddr dd ?
	zombie	_4ARGS ?
	mmu_read_byte _1ARGS ?
	mmu_read_word _1ARGS ?
	mmu_read_dword _1ARGS ?
	mmu_write_byte _2ARGS ?
	mmu_write_word _2ARGS ?
	mmu_write_dword _2ARGS ?
	mmu_get _1ARGS ?
	mmu_translate_address _1ARGS ?
	
.code


ADD_CYCLES MACRO num
	add cycle,num*200h
ENDM


	

SHIFT_OPERAND MACRO reg
	LOCAL shift_amt_reg
	
	mov ebp,60h
	and ebp,eax		; Get shift type
		
	.if ebp == 0
		shl reg,cl
	.elseif ebp == 20h
		shr reg,cl
	.elseif ebp == 40h
		sar reg,cl
	.else
		ror reg,cl
	.endif
ENDM



; Update all flags		
UPDATE_FLAGS MACRO 
	setnz al
	setnc ah
	setns bl
	setno bh

	setbe dl
	setl  dh
	setle cl

	shl ebx,16
	shl ecx,16
	or eax,ebx
	or edx,ecx
	
	mov dword ptr [flags],eax
	mov dword ptr [flags+4],edx
ENDM



; Only update N and Z + derived flags like HI, BE etc
UPDATE_NZ MACRO 
	setnz al
	;setnc ah
	setns bl
	mov bh,[flags+3]

	setbe dl
	cmp bl,bh

	setne dh
	mov cl,al
	xor cl,1
	or cl,dh

	shl ebx,16
	shl ecx,16
	or eax,ebx
	or edx,ecx
	
	mov dword ptr [flags],eax
	mov dword ptr [flags+4],edx
ENDM



LibMain PROC hInstDLL:DWORD, reason:DWORD, unused:DWORD

        .if reason == 1		;DLL_PROCESS_ATTACH
        	mov eax,hInstDLL
        	mov hInst,eax
		mov eax,TRUE
		ret

        .elseif reason == 0	;DLL_PROCESS_DETACH
        .elseif reason == 2	;DLL_THREAD_ATTACH
        .elseif reason == 3	;DLL_THREAD_DETACH
        .endif

        mov eax,FALSE
        ret
LibMain ENDP


inpb_null PROC
	ret
inpb_null ENDP
.data
inpb_handlers dd 256 DUP(OFFSET inpb_null)
.code

outpb_null PROC dat:DWORD
	ret
outpb_null ENDP
.data
outpb_handlers dd 256 DUP(OFFSET outpb_null)
.code	


op_and:
	and esi,edi
	shr eax,12
	and eax,0Fh
	mov [r0+eax*4],esi
	ADD_CYCLES 1
	ret
op_ands:
	mov ebp,eax
	and esi,edi
	UPDATE_NZ
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],esi
	ADD_CYCLES 1
	ret
op_eor:
	xor esi,edi
	shr eax,12
	and eax,0Fh
	mov [r0+eax*4],esi
	ADD_CYCLES 1
	ret
op_eors:
	mov ebp,eax
	xor esi,edi
	UPDATE_NZ
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],esi
	ADD_CYCLES 1
	ret
op_sub:
	sub esi,edi
	shr eax,12
	and eax,0Fh
	mov [r0+eax*4],esi
	ADD_CYCLES 1
	ret
op_subs:
	mov ebp,eax
	sub esi,edi
	UPDATE_FLAGS
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],esi
	ADD_CYCLES 1
	ret
op_rsb:
	sub edi,esi
	shr eax,12
	and eax,0Fh
	mov [r0+eax*4],edi
	ADD_CYCLES 1
	ret
op_rsbs:
	mov ebp,eax
	sub edi,esi
	UPDATE_FLAGS
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],edi
	ADD_CYCLES 1
	ret
op_add:
	shr eax,12
	add esi,edi
	and eax,0Fh
	mov [r0+eax*4],esi
	ADD_CYCLES 1
	ret
op_adds:
	mov ebp,eax
	add esi,edi
	UPDATE_FLAGS
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],esi
	ADD_CYCLES 1
	ret
op_adc:
	shr eax,12
	;adc esi,edi
	cmp byte ptr [flags+1],1	; Set x86 carry is ARM7 carry is set
	;cmc
	adc esi,edi
	and eax,0Fh
	mov [r0+eax*4],esi
	ADD_CYCLES 1
	ret
op_adcs:
	mov ebp,eax
	cmp byte ptr [flags+1],1	; Set x86 carry if ARM7 carry is set
	adc esi,edi
	UPDATE_FLAGS
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],esi
	ADD_CYCLES 1
	ret
op_sbc:
	shr eax,12
	cmp byte ptr [flags+1],1	; Set x86 carry is ARM7 carry is set
	sbb esi,edi
	and eax,0Fh
	mov [r0+eax*4],esi
	ADD_CYCLES 1
	ret
op_sbcs:
	mov ebp,eax
	cmp byte ptr [flags+1],1	; Set x86 carry if ARM7 carry is set
	sbb esi,edi
	UPDATE_FLAGS
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],esi
	ADD_CYCLES 1
	ret
op_rsc:
	shr eax,12
	cmp byte ptr [flags+1],1	; Set x86 carry is ARM7 carry is set
	sbb edi,esi
	and eax,0Fh
	mov [r0+eax*4],edi
	ADD_CYCLES 1
	ret
op_rscs:
	mov ebp,eax
	cmp byte ptr [flags+1],1	; Set x86 carry if ARM7 carry is set
	sbb edi,esi
	UPDATE_FLAGS
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],edi
	ADD_CYCLES 1
	ret

op_tst:
;	ret
op_tsts:
	mov ebp,eax
	and esi,edi
	UPDATE_NZ
	ADD_CYCLES 1
	ret
op_teq:
;	ret
op_teqs:
	mov ebp,eax
	xor esi,edi
	UPDATE_NZ
	ADD_CYCLES 1
	ret
op_cmp:
;	ret
op_cmps:
	mov ebp,eax
	sub esi,edi
	UPDATE_FLAGS
	ADD_CYCLES 1
	ret
op_cmn:
;	ret
op_cmns:
	mov ebp,eax
	add esi,edi
	UPDATE_FLAGS
	ADD_CYCLES 1
	ret
op_orr:
	or esi,edi
	shr eax,12
	and eax,0Fh
	mov [r0+eax*4],esi
	ADD_CYCLES 1
	ret
op_orrs:
	mov ebp,eax
	or esi,edi
	UPDATE_NZ
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],esi
	ADD_CYCLES 1
	ret
op_mov:
	shr eax,12
	and eax,0Fh
	.if eax==15
		cmp edi,[r0+56]
		jz @@not_mov_pc_lr
			add edi,4
		@@not_mov_pc_lr:
		mov physaddr,-1
	.endif
	mov [r0+eax*4],edi
	ADD_CYCLES 1
	ret
op_movs:
	mov ebp,eax
	or edi,edi
	UPDATE_NZ
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],edi
	ADD_CYCLES 1
	ret
op_bic:
	not edi
	and esi,edi
	shr eax,12
	and eax,0Fh
	mov [r0+eax*4],esi
	ADD_CYCLES 1
	ret
op_bics:
	not edi
	mov ebp,eax
	and esi,edi
	UPDATE_NZ
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],esi
	ADD_CYCLES 1
	ret
op_mvn:
	shr eax,12
	and eax,0Fh
	xor edi,-1
	.if eax==15
		cmp edi,[r0+56]
		jz @@not_mvn_pc_lr
			add edi,4
		@@not_mvn_pc_lr:
		mov physaddr,-1
	.endif
	mov [r0+eax*4],edi
	ADD_CYCLES 1
	ret
op_mvns:
	mov ebp,eax
	xor edi,-1
	UPDATE_NZ
	shr ebp,12
	and ebp,0Fh
	mov [r0+ebp*4],edi
	ADD_CYCLES 1
	ret
	


data_proc_ops:
	dd OFFSET op_and
	dd OFFSET op_ands	
	dd OFFSET op_eor
	dd OFFSET op_eors
	dd OFFSET op_sub
	dd OFFSET op_subs
	dd OFFSET op_rsb
	dd OFFSET op_rsbs
	dd OFFSET op_add
	dd OFFSET op_adds
	dd OFFSET op_adc
	dd OFFSET op_adcs
	dd OFFSET op_sbc
	dd OFFSET op_sbcs
	dd OFFSET op_rsc
	dd OFFSET op_rscs
	dd OFFSET op_tst
	dd OFFSET op_tsts
	dd OFFSET op_teq
	dd OFFSET op_teqs
	dd OFFSET op_cmp
	dd OFFSET op_cmps
	dd OFFSET op_cmn
	dd OFFSET op_cmns
	dd OFFSET op_orr
	dd OFFSET op_orrs
	dd OFFSET op_mov
	dd OFFSET op_movs
	dd OFFSET op_bic
	dd OFFSET op_bics
	dd OFFSET op_mvn
	dd OFFSET op_mvns

hword_xfer_00:
	ret
hword_xfer_01:	; STRH Rd,[Rn],-disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	invoke mmu_write_word,ebx,edx
	ADD_CYCLES 4
	ret
	
hword_xfer_02:	; STRSB Rd,[Rn],-disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	invoke mmu_write_byte,ebx,edx
	ADD_CYCLES 4
	ret
	
hword_xfer_03:	; STRSH Rd,[Rn],-disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	invoke mmu_write_word,ebx,edx
	ADD_CYCLES 4
	ret
	
hword_xfer_04:
	ret
hword_xfer_05:	; LDRH Rd,[Rn],-disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	push edx
	invoke mmu_read_word,ebx
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
hword_xfer_06:	; LDRSB Rd,[Rn],-disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	push edx
	invoke mmu_read_byte,ebx
	movsx eax,al
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret

hword_xfer_07:	; LDRSH Rd,[Rn],-disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	push edx
	invoke mmu_read_byte,ebx
	movsx eax,ax
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
hword_xfer_08:
	ret
hword_xfer_09:	; STRH Rd,[Rn],-disp!
	mov ebx,eax
	mov edx,eax
	xor esi,-1
	shr ebx,16
	shr edx,12
	inc esi

	and ebx,0Fh
	and edx,0Fh
	push ebx
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add esi,ebx
	push esi
	invoke mmu_write_word,ebx,edx
	pop esi
	pop ebx
	mov [r0+ebx*4],esi
	ADD_CYCLES 4
	ret

hword_xfer_0A:	; STRHSB Rd,[Rn],-disp!
	mov ebx,eax
	mov edx,eax
	xor esi,-1
	shr ebx,16
	shr edx,12
	inc esi
	and ebx,0Fh
	and edx,0Fh
	push ebx
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add esi,ebx
	push esi
	invoke mmu_write_byte,ebx,edx
	pop esi
	pop ebx
	mov [r0+ebx*4],esi
	ADD_CYCLES 4
	ret
	
hword_xfer_0B:	; STRSH Rd,[Rn],-disp!
	mov ebx,eax
	mov edx,eax
	xor esi,-1
	shr ebx,16
	shr edx,12
	inc esi
	and ebx,0Fh
	and edx,0Fh
	push ebx
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add esi,ebx
	push esi
	invoke mmu_write_word,ebx,edx
	pop esi
	pop ebx
	mov [r0+ebx*4],esi
	ADD_CYCLES 4
	ret
	
hword_xfer_0C:
	ret
hword_xfer_0D:	; LDRH Rd,[Rn],-disp!
	ret
hword_xfer_0E:	; LDRSB Rd,[Rn],-disp!
	ret
hword_xfer_0F:	; LDRSH Rd,[Rn],-disp!
	ret
hword_xfer_10:
	ret
hword_xfer_11:	; STRH Rd,[Rn],+disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	invoke mmu_write_word,ebx,edx
	ADD_CYCLES 4
	ret

hword_xfer_12:
	ret
hword_xfer_13:
	ret
hword_xfer_14:
	ret
hword_xfer_15:	; LDRH Rd,[Rn],+disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	push edx
	invoke mmu_read_word,ebx
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret

hword_xfer_16:
	ret
hword_xfer_17:
	ret
hword_xfer_18:
	ret
hword_xfer_19:	; STRH Rd,[Rn],+disp!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	push ebx
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add esi,ebx
	push esi
	invoke mmu_write_word,ebx,edx
	pop esi
	pop ebx
	mov [r0+ebx*4],esi
	ADD_CYCLES 4
	ret
	
hword_xfer_1A:
	ret
hword_xfer_1B:
	ret
hword_xfer_1C:
	ret
hword_xfer_1D:
	ret
hword_xfer_1E:
	ret
hword_xfer_1F:
	ret
hword_xfer_20:
	ret
hword_xfer_21:
	ret
hword_xfer_22:
	ret
hword_xfer_23:
	ret
hword_xfer_24:
	ret
hword_xfer_25:
	ret
hword_xfer_26:
	ret
hword_xfer_27:
	ret
hword_xfer_28:
	ret
hword_xfer_29:
	ret
hword_xfer_2A:
	ret
hword_xfer_2B:
	ret
hword_xfer_2C:
	ret
hword_xfer_2D:
	ret
hword_xfer_2E:
	ret
hword_xfer_2F:
	ret
hword_xfer_30:
	ret
hword_xfer_31:	; STRH Rd,[Rn,+disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add esi,ebx

	;pusha
	;invoke wsprintf,ADDR szBuffer,ADDR szFormat3,edx,esi
	;invoke lstrlen,ADDR szBuffer
	;invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
	;popa
	
	invoke mmu_write_word,esi,edx
	ADD_CYCLES 4
	ret
	
hword_xfer_32:	; STRSB Rd,[Rn,+disp]
	ret
hword_xfer_33:	; STRSH Rs,[Rn,+disp]
	ret
hword_xfer_34:
	ret
hword_xfer_35:	; LDRH Rd,[Rn,+disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	add esi,ebx
	push edx
	invoke mmu_read_word,esi
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret

hword_xfer_36:	; LDRSB Rd,[Rn,+disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	add esi,ebx
	push edx
	invoke mmu_read_byte,esi
	movsx eax,al
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
hword_xfer_37:	; LDRSH Rd,[Rn,+disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	add esi,ebx
	push edx
	invoke mmu_read_word,esi
	movsx eax,ax
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
hword_xfer_38:
	ret
hword_xfer_39:	; LDRH Rd,[Rn,+disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov edi,[r0+ebx*4]
	add esi,edi
	push edx
	mov [r0+ebx*4],esi
	invoke mmu_read_word,esi
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
hword_xfer_3A:	; LDRSB Rd,[Rn,+disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov edi,[r0+ebx*4]
	add esi,edi
	push edx
	mov [r0+ebx*4],esi
	invoke mmu_read_byte,esi
	movsx eax,al
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
hword_xfer_3B:	; LDRSB Rd,[Rn,+disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov edi,[r0+ebx*4]
	add esi,edi
	push edx
	mov [r0+ebx*4],esi
	invoke mmu_read_word,esi
	movsx eax,ax
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
hword_xfer_3C:
	ret
hword_xfer_3D:
	ret
hword_xfer_3E:
	ret
hword_xfer_3F:
	ret	
	
hword_xfer:
	dd OFFSET hword_xfer_00
	dd OFFSET hword_xfer_01
	dd OFFSET hword_xfer_02
	dd OFFSET hword_xfer_03
	dd OFFSET hword_xfer_04
	dd OFFSET hword_xfer_05
	dd OFFSET hword_xfer_06
	dd OFFSET hword_xfer_07
	dd OFFSET hword_xfer_08
	dd OFFSET hword_xfer_09
	dd OFFSET hword_xfer_0A
	dd OFFSET hword_xfer_0B
	dd OFFSET hword_xfer_0C
	dd OFFSET hword_xfer_0D
	dd OFFSET hword_xfer_0E
	dd OFFSET hword_xfer_0F
	dd OFFSET hword_xfer_10
	dd OFFSET hword_xfer_11
	dd OFFSET hword_xfer_12
	dd OFFSET hword_xfer_13
	dd OFFSET hword_xfer_14
	dd OFFSET hword_xfer_15
	dd OFFSET hword_xfer_16
	dd OFFSET hword_xfer_17
	dd OFFSET hword_xfer_18
	dd OFFSET hword_xfer_19
	dd OFFSET hword_xfer_1A
	dd OFFSET hword_xfer_1B
	dd OFFSET hword_xfer_1C
	dd OFFSET hword_xfer_1D
	dd OFFSET hword_xfer_1E
	dd OFFSET hword_xfer_1F
	dd OFFSET hword_xfer_20
	dd OFFSET hword_xfer_21
	dd OFFSET hword_xfer_22
	dd OFFSET hword_xfer_23
	dd OFFSET hword_xfer_24
	dd OFFSET hword_xfer_25
	dd OFFSET hword_xfer_26
	dd OFFSET hword_xfer_27
	dd OFFSET hword_xfer_28
	dd OFFSET hword_xfer_29
	dd OFFSET hword_xfer_2A
	dd OFFSET hword_xfer_2B
	dd OFFSET hword_xfer_2C
	dd OFFSET hword_xfer_2D
	dd OFFSET hword_xfer_2E
	dd OFFSET hword_xfer_2F
	dd OFFSET hword_xfer_30
	dd OFFSET hword_xfer_31
	dd OFFSET hword_xfer_32
	dd OFFSET hword_xfer_33
	dd OFFSET hword_xfer_34
	dd OFFSET hword_xfer_35
	dd OFFSET hword_xfer_36
	dd OFFSET hword_xfer_37
	dd OFFSET hword_xfer_38
	dd OFFSET hword_xfer_39
	dd OFFSET hword_xfer_3A
	dd OFFSET hword_xfer_3B
	dd OFFSET hword_xfer_3C
	dd OFFSET hword_xfer_3D
	dd OFFSET hword_xfer_3E
	dd OFFSET hword_xfer_3F
	

data_xfer_00:	; STR Rd,[Rn],-imm
	mov ebx,eax
	mov edx,eax
	;mov ecx,eax
	shr ebx,16
	shr edx,12
	;and ecx,0FFFh
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	invoke mmu_write_dword,ebx,edx
	ADD_CYCLES 4
	ret

data_xfer_01:	; LDR Rd,[Rn],-imm
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	push edx
	invoke mmu_read_dword,ebx
	pop edx
	ADD_CYCLES 4
	mov [r0+edx*4],eax
	ret
	
data_xfer_02:	; STR Rd,[Rn],-imm!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	xor ecx,-1
	and edx,0Fh
	inc ecx			; xor -1 / inc = neg
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add ecx,esi
	mov [r0+ebx*4],ecx
	invoke mmu_write_dword,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_03:	; LDR Rd,[Rn],-imm!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	xor ecx,-1
	and edx,0Fh
	inc ecx

	mov esi,[r0+ebx*4]
	push edx
	add ecx,esi
	mov [r0+ebx*4],ecx
	invoke mmu_read_dword,esi
	pop edx
	ADD_CYCLES 4
	mov [r0+edx*4],eax
	ret
	
data_xfer_04:	; STRB Rd,[Rn],-disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	invoke mmu_write_byte,ebx,edx
	ADD_CYCLES 4
	ret
	
data_xfer_05:	; LDRB Rd,[Rn],-disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	push edx
	invoke mmu_read_byte,ebx
	pop edx
	and eax,0FFh
	ADD_CYCLES 4
	mov [r0+edx*4],eax
	ret
	
data_xfer_06:	; STRB Rd,[Rn],-disp!
	mov ebx,eax
	mov edx,eax
	xor ecx,-1
	shr ebx,16
	shr edx,12
	inc ecx

	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add ecx,esi
	mov [r0+ebx*4],ecx
	invoke mmu_write_byte,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_07:	; LDRB Rd,[Rn],-disp!
	mov ebx,eax
	mov edx,eax
	xor ecx,-1
	shr ebx,16
	shr edx,12
	inc ecx

	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	push edx
	add ecx,esi

	mov [r0+ebx*4],ecx
	invoke mmu_read_byte,esi
	pop edx
	and eax,0FFh
	ADD_CYCLES 4
	mov [r0+edx*4],eax
	ret
	
data_xfer_08:	; STR Rd,[Rn],+disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	invoke mmu_write_dword,ebx,edx
	ADD_CYCLES 4
	ret
	
data_xfer_09:	; LDR Rd,[Rn],+disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	push edx
	invoke mmu_read_dword,ebx
	pop edx
	ADD_CYCLES 4
	mov [r0+edx*4],eax
	ret
	
data_xfer_0A:	; STR Rd,[Rn],+imm!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add ecx,esi
	mov [r0+ebx*4],ecx
	invoke mmu_write_dword,esi,edx
	ADD_CYCLES 4
	ret

data_xfer_0B:	; LDR Rd,[Rn],+disp!
	mov ebx,eax
	mov edx,eax
	;mov ecx,eax
	shr ebx,16
	shr edx,12
	;and ecx,0FFFh
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	add ecx,esi
	push edx
	mov [r0+ebx*4],ecx
	invoke mmu_read_dword,esi
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret

data_xfer_0C:	; STRB Rd,[Rn],+disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	mov edx,[r0+edx*4]
	invoke mmu_write_byte,ebx,edx
	ADD_CYCLES 4
	ret
	
data_xfer_0D:	; LDRB Rd,[Rn],+disp
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	push edx
	invoke mmu_read_byte,ebx
	pop edx
	and eax,0FFh
	ADD_CYCLES 4
	mov [r0+edx*4],eax
	ret

data_xfer_0E:	; STRB Rd,[Rn],+disp!
	mov ebx,eax
	mov edx,eax
	;mov ecx,eax
	shr ebx,16
	shr edx,12
	;and ecx,0FFFh
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add ecx,esi
	mov [r0+ebx*4],ecx
	invoke mmu_write_byte,esi,edx
	ADD_CYCLES 4
	ret

data_xfer_0F:	; LDRB Rd,[Rn],+imm!
	mov ebx,eax
	mov edx,eax
	;mov ecx,eax
	shr ebx,16
	shr edx,12
	;and ecx,0FFFh
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	add ecx,esi
	push edx
	mov [r0+ebx*4],ecx
	invoke mmu_read_byte,ecx
	pop edx
	and eax,0FFh
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
data_xfer_10:	; STR Rd,[Rn,-disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	sub esi,ecx
	invoke mmu_write_dword,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_11:	; LDR Rd,[Rn,-disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	sub ebx,ecx
	push edx
	invoke mmu_read_dword,ebx
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
data_xfer_12:	; STR Rd,[Rn,-disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	sub esi,ecx
	mov [r0+ebx*4],esi
	invoke mmu_write_dword,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_13:	; LDR Rd,[Rn,-disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	push edx
	sub esi,ecx
	mov [r0+ebx*4],esi
	invoke mmu_read_dword,esi
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret

data_xfer_14:	; STRB Rd,[Rn,-disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	sub esi,ecx
	invoke mmu_write_byte,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_15:	; LDRB Rd,[Rn,-disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	sub ebx,ecx
	push edx
	invoke mmu_read_byte,ebx
	and eax,0FFh
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
data_xfer_16:	; STRB Rd,[Rn,-disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	sub esi,ecx
	mov [r0+ebx*4],esi
	invoke mmu_write_byte,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_17:	; LDRB Rd,[Rn,-disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	push edx
	sub esi,ecx
	mov [r0+ebx*4],esi
	invoke mmu_read_byte,esi
	and eax,0FFh
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
data_xfer_18:	; STR Rd,[Rn,+disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add esi,ecx
	;mov [r0+ebx*4],ecx
	invoke mmu_write_dword,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_19:	; LDR Rd,[Rn,+disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	add ecx,ebx
	push edx
	invoke mmu_read_dword,ecx
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret

data_xfer_1A:	; STR Rd,[Rn,+disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add esi,ecx
	mov [r0+ebx*4],esi
	invoke mmu_write_dword,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_1B:	; LDR Rd,[Rn,+disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	add esi,ecx
	push edx
	mov [r0+ebx*4],esi
	invoke mmu_read_dword,esi
	pop edx
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret
	
data_xfer_1C:	; STRB Rd,[Rn,+disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add esi,ecx
	invoke mmu_write_byte,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_1D:	; LDRB Rd,[Rn,+disp]
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov ebx,[r0+ebx*4]
	add ecx,ebx
	push edx
	invoke mmu_read_byte,ecx
	pop edx
	and eax,0FFh
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret

data_xfer_1E:	; STRB Rd,[Rn,+disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	mov edx,[r0+edx*4]
	add esi,ecx
	mov [r0+ebx*4],esi

	invoke mmu_write_byte,esi,edx
	ADD_CYCLES 4
	ret
	
data_xfer_1F:	; LDRB Rd,[Rn,+disp]!
	mov ebx,eax
	mov edx,eax
	shr ebx,16
	shr edx,12
	and ebx,0Fh
	and edx,0Fh
	mov esi,[r0+ebx*4]
	add esi,ecx
	push edx
	mov [r0+ebx*4],esi
	invoke mmu_read_byte,esi
	pop edx
	and eax,0FFh
	mov [r0+edx*4],eax
	ADD_CYCLES 4
	ret

data_xfer:
	dd OFFSET data_xfer_00
	dd OFFSET data_xfer_01
	dd OFFSET data_xfer_02
	dd OFFSET data_xfer_03
	dd OFFSET data_xfer_04
	dd OFFSET data_xfer_05
	dd OFFSET data_xfer_06
	dd OFFSET data_xfer_07
	dd OFFSET data_xfer_08
	dd OFFSET data_xfer_09
	dd OFFSET data_xfer_0A
	dd OFFSET data_xfer_0B
	dd OFFSET data_xfer_0C
	dd OFFSET data_xfer_0D
	dd OFFSET data_xfer_0E
	dd OFFSET data_xfer_0F
	dd OFFSET data_xfer_10
	dd OFFSET data_xfer_11
	dd OFFSET data_xfer_12
	dd OFFSET data_xfer_13
	dd OFFSET data_xfer_14
	dd OFFSET data_xfer_15
	dd OFFSET data_xfer_16
	dd OFFSET data_xfer_17
	dd OFFSET data_xfer_18
	dd OFFSET data_xfer_19
	dd OFFSET data_xfer_1A
	dd OFFSET data_xfer_1B
	dd OFFSET data_xfer_1C
	dd OFFSET data_xfer_1D
	dd OFFSET data_xfer_1E
	dd OFFSET data_xfer_1F


block_xfer_00:
	ret
block_xfer_01:
	ret
block_xfer_02:
	ret
block_xfer_03:
	ret
block_xfer_04:
	ret
block_xfer_05:
	ret
block_xfer_06:
	ret
block_xfer_07:
	ret
block_xfer_08:
	ret
block_xfer_09:
	ret
block_xfer_0A:
	ret

block_xfer_0B:	; LDMFD Rn!,{...}

	mov ebx,eax
	mov esi,OFFSET r0
	shr ebx,16
	mov ecx,0
	and ebx,0Fh		
	mov edx,eax
	mov ebp,[esi+ebx*4]	; get Rn
	@@block_xfer_0B_loop:
		shr edx,1
		jnc @@block_xfer_0B_no_xfer
			push ebx
			push ecx
			push edx
			push ebp
			push esi
			invoke mmu_read_dword,ebp
			pop esi
			pop ebp
			pop edx
			pop ecx
			pop ebx
			mov [esi+ecx*4],eax
			add ebp,4
		@@block_xfer_0B_no_xfer:
		inc ecx
		cmp ecx,16
		jne @@block_xfer_0B_loop
	mov [esi+ebx*4],ebp
	ADD_CYCLES 8
	ret

block_xfer_0C:
	ret
block_xfer_0D:
	ret
block_xfer_0E:
	ret
block_xfer_0F:
	ret
block_xfer_10:
	ret
block_xfer_11:
	ret
block_xfer_12:	; STMFD Rn!,{...}

	add r15,4
	mov ebx,eax
	mov esi,OFFSET r0
	shr ebx,16
	mov ecx,15
	and ebx,0Fh		
	mov ebp,[esi+ebx*4]	; get Rn
	shl eax,16
	@@block_xfer_12_loop:
		shl eax,1
		jnc @@block_xfer_12_no_xfer
			mov edx,[esi+ecx*4]
			sub ebp,4
			pusha
			invoke mmu_write_dword,ebp,edx
			popa
		@@block_xfer_12_no_xfer:
		dec ecx
		;cmp ecx,16
		jns @@block_xfer_12_loop
	mov [esi+ebx*4],ebp
	sub r15,4
	ADD_CYCLES 8
	ret

block_xfer_13:
	ret
block_xfer_14:
	ret
block_xfer_15:
	ret
block_xfer_16:
	ret
block_xfer_17:
	ret
block_xfer_18:
	ret
block_xfer_19:
	ret
block_xfer_1A:
	ret
block_xfer_1B:
	ret
block_xfer_1C:
	ret
block_xfer_1D:
	ret
block_xfer_1E:
	ret
block_xfer_1F:
	ret

	
block_xfer:
	dd OFFSET block_xfer_00
	dd OFFSET block_xfer_01
	dd OFFSET block_xfer_02
	dd OFFSET block_xfer_03
	dd OFFSET block_xfer_04
	dd OFFSET block_xfer_05
	dd OFFSET block_xfer_06
	dd OFFSET block_xfer_07
	dd OFFSET block_xfer_08
	dd OFFSET block_xfer_09
	dd OFFSET block_xfer_0A
	dd OFFSET block_xfer_0B
	dd OFFSET block_xfer_0C
	dd OFFSET block_xfer_0D
	dd OFFSET block_xfer_0E
	dd OFFSET block_xfer_0F
	dd OFFSET block_xfer_10
	dd OFFSET block_xfer_11
	dd OFFSET block_xfer_12
	dd OFFSET block_xfer_13
	dd OFFSET block_xfer_14
	dd OFFSET block_xfer_15
	dd OFFSET block_xfer_16
	dd OFFSET block_xfer_17
	dd OFFSET block_xfer_18
	dd OFFSET block_xfer_19
	dd OFFSET block_xfer_1A
	dd OFFSET block_xfer_1B
	dd OFFSET block_xfer_1C
	dd OFFSET block_xfer_1D
	dd OFFSET block_xfer_1E
	dd OFFSET block_xfer_1F
	
	
	
type_0:
	mov ebx,eax
	and ebx,90h
	cmp ebx,90h
	je @@hword_op_or_mul
	mov ebx,eax
	mov esi,eax
	mov edi,eax
	shr ebx,20		; Get opcode
	shr esi,16		; Get Rn
	and edi,0Fh		; Get Rm
	and esi,0Fh
	and ebx,01Fh
	mov edi,[r0+edi*4]
	mov esi,[r0+esi*4]

	.if (eax & 10h)			; Shift amount specified in register
		mov ecx,eax
		shr ecx,8
		and ecx,0Fh
		mov ecx,[r0+ecx*4]
		SHIFT_OPERAND edi
	.else				; Shift amount specified in immediate value	
		test eax,0F80h		
		jz @@type_0_no_shift	; Only perform the shift if imm>0
			mov ecx,eax
			shr ecx,7	; Get Rm shift
			SHIFT_OPERAND edi
		@@type_0_no_shift:
	.endif
	
	jmp dword ptr [data_proc_ops + ebx*4]

	@@hword_op_or_mul:
	;mov r17,eax
	test eax,60h
	jnz @@hword_op
	test eax,1000000h
	jnz @@hword_op

	@@mul_or_mla:
	; Cond 000000 AS Rd Rn  Rs  1001 Rm
	; Rd = Rs*Rm | Rd = Rs*Rm + Rn
	mov ebx,eax
	mov ecx,eax
	mov ebp,eax

	mov edi,eax
	shr ebx,8
	and ecx,0Fh		; Rm
	push eax

	and ebx,0Fh		; Rs
	mov eax,[r0+ecx*4]
	and ebp,300000h
	.if ebp==0

	pusha
	;mov ebx,[r0+ebx*4]
	mov ecx,edi
	shr ecx,16
	and ecx,0Fh
	invoke wsprintf,ADDR szBuffer,ADDR szFormat9,ecx,eax,ebx
	invoke lstrlen,ADDR szBuffer
	invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
	popa
	
		mul dword ptr [r0+ebx*4]		

		mov esi,eax
	.elseif ebp==100000h
		mul dword ptr [r0+ebx*4]
		mov esi,eax
		UPDATE_NZ
	.elseif ebp==200000h
		shr edi,12
		mul dword ptr [r0+ebx*4]
		and edi,0Fh
		add eax,[r0+edi*4]
		mov esi,eax
	.else
		shr edi,12
		mul dword ptr [r0+ebx*4]
		and edi,0Fh
		add eax,[r0+edi*4]
		mov esi,eax
		UPDATE_NZ
	.endif
	pop edx
	shr edx,16
	and edx,0Fh
	mov [r0+edx*4],esi
	

	@@hword_op:	
	mov ebx,eax
	mov ecx,eax
	shr ebx,18		
	shr ecx,5		
	and ebx,07Ch		; ebx = PU0WL00
	and ecx,3		; ecx = SH
	mov edx,ebx
	or ebx,ecx		; ebx = PU0WLSH
	shr edx,1		; edx = 0PU0WL0
	and ebx,0Fh		; ebx = 000WLSH
	and edx,30h		; edx = 0PU0000 
	or ebx,edx		; ebx = 0PUWLSH
	test eax,400000h
	jz @@hword_op_reg_offs
		mov esi,eax
		mov edi,eax
		and esi,0Fh
		shr edi,4
		and edi,0F0h
		or esi,edi	; esi = 8bit offset
		jmp dword ptr [hword_xfer + ebx*4]
	@@hword_op_reg_offs:
		mov edi,eax
		and edi,0Fh
		mov esi,[r0+edi*4] ; esi = offset register	
		jmp dword ptr [hword_xfer + ebx*4]
		
type_1:
	mov ebx,eax
	mov esi,eax
	mov edi,eax
	shr ebx,20		; Get opcode
	shr esi,16		; Get Rn
	and edi,0FFh		; Get Op2
	and esi,0Fh
	and ebx,01Fh
	mov esi,[r0+esi*4]

	test eax,0F80h
	jz @@type_1_no_rot
		mov ecx,eax
		shr ecx,8		; Get Op2 shift
		and ecx,0Fh
		add ecx,ecx
		ror edi,cl
	@@type_1_no_rot:
	jmp dword ptr [data_proc_ops + ebx*4]

; Type 2 - LDR/STR Rd,[Rn],imm	
type_2:
	mov ebx,eax
	mov ecx,eax
	shr ebx,20
	and ecx,0FFFh
	and ebx,01Fh
	
	;pusha
	;invoke wsprintf,ADDR szBuffer,ADDR szFormat2,ebx,r15
	;invoke lstrlen,ADDR szBuffer
	;invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
	;popa
	jmp dword ptr [data_xfer + ebx*4]

; Type 3 - LDR/STR Rd,[Rn],Rm	
type_3:
	mov ebx,eax
	mov edi,eax
	shr ebx,20
	and edi,0Fh
	and ebx,01Fh
	mov edi,[r0+edi*4]

	test eax,0F80h
	jz @@type_3_no_shift
		mov ecx,eax
		shr ecx,7		; Get Rm shift
		SHIFT_OPERAND edi
	@@type_3_no_shift:
	mov ecx,edi
	
	jmp dword ptr [data_xfer + ebx*4]


; Type 4 - Block transfer
type_4:
	mov ebx,eax
	shr ebx,20
	and ebx,01Fh
	jmp dword ptr [block_xfer + ebx*4]
	
; Type 5 - Branch
type_5:
	test eax,1000000h
	jz @@no_link
		mov ebx,r15
		;add ebx,4
		mov r14,ebx
	@@no_link:
	and eax,0FFFFFFh	; Get 24-bit offset
	shl eax,8
	sar eax,6		; Shift left by 2 and sign-extend
	add eax,4		

	mov physaddr,-1

	;mov ebx,r15
	add r15,eax
	;.if ebx == r14
	;	pusha
	;	invoke wsprintf,ADDR szBuffer,ADDR szFormat4,r14,r15
	;	invoke lstrlen,ADDR szBuffer
	;	invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
	;	popa
	;.endif
	
	ADD_CYCLES 4
	ret

type_6:
	ret
type_7:
	ret
	

instr_decoder:
	dd OFFSET type_0
	dd OFFSET type_1
	dd OFFSET type_2
	dd OFFSET type_3
	dd OFFSET type_4
	dd OFFSET type_5
	dd OFFSET type_6
	dd OFFSET type_7



cpu_init PROC commlink:DWORD	
	mov eax,commlink
	mov zombie,eax

	invoke GetStdHandle,STD_OUTPUT_HANDLE
	mov hout,eax
	
	ret
cpu_init ENDP


cpu_reset PROC
	pusha

	invoke zombie,ASK,MMU_READ_BYTE_CALLBACK,0,0
	mov mmu_read_byte,eax
	invoke zombie,ASK,MMU_READ_WORD_CALLBACK,0,0
	mov mmu_read_word,eax
	invoke zombie,ASK,MMU_READ_DWORD_CALLBACK,0,0
	mov mmu_read_dword,eax

	invoke zombie,ASK,MMU_WRITE_BYTE_CALLBACK,0,0
	mov mmu_write_byte,eax
	invoke zombie,ASK,MMU_WRITE_WORD_CALLBACK,0,0
	mov mmu_write_word,eax
	invoke zombie,ASK,MMU_WRITE_DWORD_CALLBACK,0,0
	mov mmu_write_dword,eax

	invoke zombie,ASK,MMU_TRANSLATE_ADDRESS_CALLBACK,0,0
	mov mmu_translate_address,eax
	
	popa
	ret
cpu_reset ENDP


cpu_close PROC
	ret
cpu_close ENDP


cpu_config PROC hWnd:DWORD
	ret
cpu_config ENDP


cpu_set_io_handler PROC mode:DWORD,port:DWORD,pfhandler:DWORD
	push ebx

	mov eax,port
	mov ebx,pfhandler
	
	.if mode==INPUT	
		mov [inpb_handlers + eax*4],ebx
	.else
		mov [outpb_handlers + eax*4],ebx
	.endif
	
	pop ebx
	ret
cpu_set_io_handler ENDP


cpu_get PROC what:DWORD
	.if what == CPU_CYCLE_COUNT
		mov eax,cycle
	.elseif what == CPU_CYCLE_SCALING
		mov eax,200h
	.elseif what == CPU_SET_IO_CALLBACK
		mov eax,offset cpu_set_io_handler
	.elseif what == CPU_IRQ_CALLBACK
		mov eax,offset cpu_irq
	.elseif what == CPU_INPB_CALLBACK
		mov eax,offset cpu_inpb
	.elseif what == CPU_OUTPB_CALLBACK
		mov eax,offset cpu_outpb
	.elseif what == CPU_SPEED
		mov eax,speed
	.elseif what == CPU_REG_PC
		mov eax,r15
	;.elseif what == CPU_REG
	;	mov eax,param1
	;	and eax,0Fh
	;	mov eax,[r0+eax*4]
	.elseif what == PLUGIN_NAME
		mov eax,offset cpuName
	.endif
	ret
cpu_get ENDP


cpu_set PROC what:DWORD,param1:DWORD,param2:DWORD
	.if what == CPU_REG_PC
		mov eax,param1
		mov r15,eax
	.elseif what == CPU_REG_SP
		mov eax,param1
		mov r13,eax
	.elseif what == CPU_CYCLE_COUNT
		mov eax,param1
		mov cycle,eax
	.elseif what == CPU_BREAKPOINT
		mov eax,param1
		mov breakpoint,eax
	.elseif what == CPU_SPEED
		mov eax,param1
		mov speed,eax
	.elseif what == CPU_IO_HANDLER
		mov eax,param1
		and param1,0FFh
		shr eax,8
		invoke cpu_set_io_handler,eax,param1,param2
	.endif
	ret
cpu_set ENDP



cpu_outpb PROC port:DWORD,dat:DWORD
	push ecx
	push dword ptr dat
	mov eax,port
	call dword ptr [outpb_handlers + eax*4]
	pop ecx
	ret
cpu_outpb ENDP


cpu_inpb PROC port:DWORD
	push ecx
	mov eax,port
	call dword ptr [inpb_handlers + eax*4]
	pop ecx
	ret
cpu_inpb ENDP


cpu_irq PROC address:DWORD
	ret
cpu_irq ENDP


cpu_execute PROC
	ret
cpu_execute ENDP


cpu_execute_until PROC threshold:DWORD
	LOCAL lastCyc:DWORD
	
	pusha
	
	mov ecx,cycle
	mov physaddr,-1
	
	@@exe_loop:
		cmp ecx,threshold
		jge @@exe_done
		mov lastCyc,ecx

		.if physaddr == -1
			mov eax,r15
			;invoke mmu_read_dword,eax
			invoke mmu_translate_address,eax
			mov physaddr,eax
		.endif
		mov ebx,physaddr
		mov eax,[ebx]
		add physaddr,4
		
		mov ebx,eax
		and ebx,0F0000000h
		cmp ebx,0E0000000h
		je @@cond_true
		shr ebx,28
		mov edx,ebx
		and edx,1
		shr ebx,1
		cmp dl,[flags+ebx]
		jne @@cond_false
		@@cond_true:
			
			mov ebx,eax
			push ebp
			shr ebx,25
			add r15,8
			and ebx,07h
			
			call dword ptr [instr_decoder + ebx*4]
			pop ebp
			mov ecx,cycle
			sub r15,4
			cmp ecx,lastCyc
			jne @@exe_loop
			jmp @@exe_done
			
		@@cond_false:
		ADD_CYCLES 1
		add r15,4
		mov ecx,cycle
		
		; !! Temp fix !!
		cmp ecx,lastCyc
		jne @@exe_loop
		mov ecx,1
	 
	@@exe_done:
	;mov cycle,ecx
	popa
	
	mov eax,cycle
	ret
cpu_execute_until ENDP


END LibMain
