; Zombie ARM7 example
; /Mic, 2004
;
; Fire effect in bitmapped mode
;
; Assemble with goldroad (http://www.goldroad.co.uk)


@fsize 1	

@define GPU_REG0	$0000FF00
@define GPU_REG1	$0000FF01
@define SCREEN		$06000000
@define PALETTE		$06012C00




@textarea $8000000		; Start of ROM area

start

ldr sp,=$3007000		; Set stack pointer to a writeable address	

; Copy code from ROM to RAM
mov r0,ram_section
mov r1,ram_section_ends
sub r2,r1,r0			; # of bytes to move
mov r1,#$3000000		; Target address
copy_code
	ldr r3,[r0],#4
	str r3,[r1],#4
	subs r2,r2,#4
bne copy_code


mov pc,#$3000000		; Branch to RAM

@ltorg				; Dump literal pool
@endarea			; End of ROM area



ram_section
@textarea $03000000

	
b ram_entry

@include rand.asm		; Include RNG


VSync
; In:
;  -
;
; Out:
;  -
	stmfd sp!,{r0-r1}	; Save registers
	ldr r0,=GPU_REG1	; GPU register 1 holds GPU status bits
	
	VS_loop
		ldrb r1,[r0]	; Read GPU reg 1
		ands r1,r1,#80	; Check vblank bit (cleared on read)
		beq VS_loop	; Loop until set
	
	ldmfd sp!,{r0-r1}	; Restore registers
	mov pc,lr		; Return


ram_entry

ldr r0,=GPU_REG0
mov r1,#0
strb r1,[r0]			; Disable display, vblank interrupts etc


mov r3,#0
ldr r4,=PALETTE
setup_palette
  	mov r0,r3,lsl #11
  	strh r0,[r4]
  	
  	mov r1,#31
  	mov r0,r3,lsl #6
  	orr r0,r0,r1,lsl #11
  	strh r0,[r4,#64]
  
  	mov r0,#$FF00
  	mov r1,#$E0
	orr r0,r0,r1
	orr r0,r0,r3
  	strh r0,[r4,#128]
  	add r4,r4,#2
  
  	add r3,r3,#1
  	cmp r3,#32
	bne setup_palette
pal_done

; Set the remaining colors to white
ldr r0,=PALETTE
add r0,r0,#192
ldr r1,=$FFFF
mov r2,160			; 256 colors, minus the 96 already defined
set_colors_to_white
  	strh r1,[r0],#2
  	subs r2,r2,#1
	bne set_colors_to_white


ldr r0,=GPU_REG0
mov r1,#$B4
strb r1,[r0]			; Enable display, vblank interrupts, bitmapped mode 



ldr r5,=644
ldr r10,=320
ldr r11,=75840				; 75840 = 237*320

forever

  	;bl VSync	

  	mov r6,0			; x = 0
  	x_loop
  		mov r7,SCREEN
  		add r7,r7,r6
  		ldr r8,=22718		; Read-offset
  		ldr r9,=22080		; Write-offset
  		add r8,r8,r7
  		add r9,r9,r7
  		mov r7,#71		; y = 80
  		y_loop
  	  		ldrb r0,[r8]	
  	  		ldrb r1,[r8,#2]
  	  		ldrb r2,[r8,#4]
  	  		ldrb r3,[r8,r5]

   	  		add r0,r0,r1
  	  		add r0,r0,r2
  	  		add r0,r0,r3
  	  		mov r0,r0,lsr #2 	; Divide the color sum by 4

  	  		cmp r0,#0
  	  		subgt r0,r0,#1		; Subtract 1 if color average > 1
  	  		orr r0,r0,r0,lsl #8	; Copy bits 0..7 to 8..15

  	  		strh r0,[r9]		; Draw 2 pixels
  	  		strh r0,[r9,r10]	; ..2 more

  	  		add r8,r8,#640		; Jump 2 lines down
  	  		add r9,r9,#640

  	  		add r7,r7,#2
  	  		cmp r7,#239		; Done ?
  			blt y_loop

  		bl Random16		; Create a "random" number. the result is given in r0
  		add r0,r0,#28		
  		and r0,r0,#191

  		mov r1,SCREEN
  		add r1,r1,r11
  		add r1,r1,r6		; Add x
   	 	strb r0,[r1]		; Store random pixel
  
  		add r6,r6,#2
  		cmp r6,#320		; Done ?
blt x_loop


b forever		; Loop the loop..


; These two LDRs are just here because the literal
; pool will get messed up without them (for some reason).
ldr r0,=1234
ldr r1,=5678


@ltorg			; Dump literal pool
@endarea

ram_section_ends
