// Blackfin fire effect for Zombie
// /Mic, 2005
//
// bfin-elf-as -o fire.o fire.s
// bfin-elf-ld -T lnkscript fire.o
// bfin-elf-objcopy -O binary a.out fire.bin


// Address definitions for the fictive GPU
.EQU GPUREG0,	0xFF00
.EQU GPUREG1, 	0xFF01
.EQU VRAM, 	0x6000000
.EQU CRAM,	0x6012C00

// Misc. constants
.EQU READOFS, 	22718
.EQU WRITEOFS, 	22080
.EQU SCREENOFS,	75840


// Macro for loading 32-bit immediate values
.MACRO MOVIMM32 regn,imm32
	\regn.H = \imm32; \regn.L = \imm32
.ENDM


.TEXT
.ALIGN 2

.GLOBAL _start

_start:
	// Copy initialised data from ROM to RAM (132 bytes)
	MOVIMM32 P0,rng_data_rom;
	MOVIMM32 P1,data_ram;
	R0 = 33;
	LC0 = R0;				// 33 = 132/4
	LOOP copy_data LC0;
	LOOP_BEGIN copy_data;
		R0 = [P0++];
		[P1++] = R0;
		LOOP_END copy_data;
		
	P5 = 0xFF00(Z); 			// FF00 = GPU register 0		
	R0 = 0;
	B[P5] = R0;				// Disable screen


	// Setup palette
	R7 = 0;
	R6 = 0x20;
	MOVIMM32 P4,CRAM;			// Color RAM address
	
	setup_palette:
		R0 = R7;

		R0 <<= 11;
		W[P4] = R0;			// [P4] = R7 << 11

		R0 = R7;
		R1 = 0xF800(Z);
		R0 <<= 6;
		R0 = R0 | R1;
		W[P4 + 64] = R0;		// [P4 + 64] = 0z1F<<11 | R7<<6
		
		R0 = 0xFFE0(Z);
		R0 = R0 | R7;
		W[P4 + 128] = R0;		// [P4 + 128] = 0xFFE0 | R7
		
		P4 += 2;
		R7 += 1;
		CC = R7<R6;
		IF CC JUMP setup_palette;
		
		
	//Set the remaining colors to white
	MOVIMM32 P4,CRAM+192;			// We have already set 96 colors (=192 bytes)
	R7 = 0xFFFF(Z);				// 0xFFFF == full white
	R0 = 160(Z); LC0 = R0;			// 160 = 256 - 96
	LOOP set_colors_to_white LC0;
	LOOP_BEGIN set_colors_to_white;
		W[P4++] = R7;
		LOOP_END set_colors_to_white;

	R0 = 0xB4(Z);
	B[P5] = R0;				// Enable display, vblank interrupts, bitmapped mode
	

	forever:
		// Wait for vertical blank
		P4 = 0xFF01(Z);
		wait_vb:
			R1 = B[P4](Z); 
			CC = BITTST(R1,7);	// Check vblank bit
			IF !CC JUMP wait_vb;
			
	
		P5 = 0;				// X
		x_loop:
			MOVIMM32 P4,VRAM;
			P4 = P4 + P5;
			P0 = READOFS(Z);
			P1 = WRITEOFS(Z);
			P0 = P0 + P4;
			P1 = P1 + P4;
			R0 = 84(Z); LC0 = R0;	// 168 lines
			LOOP y_loop LC0;
			LOOP_BEGIN y_loop;
				// Read 4 pixels
				R0 = B[P0](Z);
				R1 = B[P0 + 2](Z);
				R2 = B[P0 + 4](Z);
				R3 = B[P0 + 644](Z);
				
				// Calculate average
				R0 = R0 + R1;
				R0 = R0 + R2;
				R0 = R0 + R3;
				R0 >>= 2;
				
				// Decrement color value if >0
				CC = R0 == 0;
				R1 = R0;
				R0 += -1;
				IF CC R0 = R1;

				// Duplicate bits 0..7 into bits 8..15
				R1 = R0;
				R1 <<= 8;
				R0 = R0 | R1;
				
				W[P1] = R0;	
				W[P1 + 320] = R0;
				
				P3 = 640(Z);
				P0 = P0 + P3;	// Move two rows down
				P1 = P1 + P3;
				
				LOOP_END y_loop;

			CALL Random;
			R5 += 28;
			R3 = 191(Z);
			R5 = R5 & R3;
			
			MOVIMM32 P4,VRAM
			MOVIMM32 P3,SCREENOFS
			P4 = P4 + P3;
			P4 = P4 + P5;		// Add X
			B[P4] = R5;		// Store random pixel
						
			P3 = 320(Z);
			P5 += 2;
			CC = P5==P3;
			IF !CC JUMP x_loop;
	
		JUMP forever;			


	// Pseudo-random number generator
	Random:
		MOVIMM32 P2,data_ram;		// Point P2 to seed1 in RAM
		P3 = P2;
		P3 += 2;			// Point P3 to the random number table
		R3 = B[P2](Z);			// R3 = seed1
		R4 = B[P2 + 1](Z);		// R4 = seed2
		P0 = R3;
		P1 = R4;
		P0 = P3 + (P0 << 1);
		P1 = P3 + (P1 << 1);
		R5 = W[P0](Z);			// Load random number
		R6 = W[P1](Z);			// ...another one
		R5 = R5 + R6;			// Add the two numbers
		W[P1] = R5;			// Store it
		R3 += 1;			// Increase seed1
		R4 += 1;			// Increase seed2
		R2 = 0x3F;			// The table contains 64 entries
		R3 = R3 & R2;
		R4 = R4 & R2;
		B[P2] = R3;
		B[P2 + 1] = R4;
		RTS;
		
	
.ALIGN 2

rng_data_rom:	
	# Seeds
	.BYTE 29
	.BYTE 63

	# 64 random 16-bit numbers for use with the random-number generator 
	.SHORT 25662,56977,36314,46707
	.SHORT 56747,47936,28467,50155
	.SHORT 54470,9126,19191,26274
	.SHORT 27205,13469,6769,48786
	.SHORT 43195,17969,58549,13097
	.SHORT 9962,26613,60724,25428
	.SHORT 44416,43884,31777,40699
	.SHORT 4280,32767,31882,41912
	.SHORT 19050,42386,38267,39411
	.SHORT 11739,36780,10921,43911
	.SHORT 7473,38901,28614,62047
	.SHORT 25094,22064,16149,14666
	.SHORT 46209,34643,30017,25842
	.SHORT 64242,3527,34102,35358
	.SHORT 2277,3678,16479,3431
	.SHORT 36135,63741,11311,0
          

.DATA
data_ram:
	seed1: .BYTE
	seed2: .BYTE
	random_tbl:
	