; A dynamic recompilating (binary code translating) MIPS R2000 emulator
; /Mic, 2004


.586p
.model flat,stdcall
option casemap:none


include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\user32.inc

includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\user32.lib


.data

	; Message enumerators
	CPU_IRQ_CALLBACK		equ 000h
	CPU_SET_IO_CALLBACK		equ 001h
	CPU_IO_HANDLER			equ 001h
	CPU_CYCLE_COUNT			equ 002h
	CPU_REG_PC			equ 003h
	CPU_REG_SP			equ 004h
	CPU_BREAKPOINT			equ 005h
	CPU_SPEED			equ 006h
	CPU_INPB_CALLBACK		equ 007h
	CPU_OUTPB_CALLBACK		equ 008h
	CPU_CYCLE_SCALING		equ 009h
	CPU_REG				equ 00Ah
	
	MMU_READ_BYTE_CALLBACK		equ 100h
	MMU_READ_WORD_CALLBACK		equ 101h
	MMU_WRITE_BYTE_CALLBACK		equ 102h
	MMU_WRITE_WORD_CALLBACK		equ 103h
	MMU_TRANSLATE_ADDRESS_CALLBACK	equ 104h
	MMU_READ_DWORD_CALLBACK		equ 106h
	MMU_WRITE_DWORD_CALLBACK	equ 107h

	GPU_CURRENT_SCANLINE		equ 200h
	GPU_SCREEN_WIDTH		equ 201h
	GPU_SCREEN_HEIGHT		equ 202h
	GPU_FRAME_COMPLETED		equ 203h
	GPU_SCREEN_PITCH		equ 204h

	ZOMBIE_WINDOW_HANDLE		equ 500h

	PLUGIN_NAME			equ 600h
	
	INPUT		equ 0
	OUTPUT		equ 1

	ASK		equ 0
	TELL		equ 1


	; Prototypes
	_1ARGS_PROTO	typedef PROTO :DWORD
	_2ARGS_PROTO	typedef PROTO :DWORD,:DWORD
	_4ARGS_PROTO	typedef PROTO :DWORD,:DWORD,:DWORD,:DWORD
	_1ARGS		typedef ptr _1ARGS_PROTO
	_2ARGS		typedef ptr _2ARGS_PROTO
	_4ARGS		typedef ptr _4ARGS_PROTO

	cpu_inpb	PROTO :DWORD
	cpu_outpb	PROTO :DWORD,:DWORD

	wsprintfA PROTO C :DWORD, :VARARG
	wsprintf equ <wsprintfA>


	; Translation cache structure
	cache_entry STRUCT
		codePtr	dd ?	; Pointer to start of x86 code
		inUse 	dd ?	; Set if cache line is in use, clear otherwise
		address dd ?	; Start address of MIPS code block
		cycles 	dd ?	; Number of cycles used by MIPS code
	cache_entry ENDS
	
	
	; Cycle count
	cycle		dd 0

	; Breakpoint address
	breakpoint	dd 0

	; Default speed for this CPU (33 MHz)
	speed		dd 2100000h	; 33 MHz

	; Name of this CPU
	cpuName		db "MIPS R2000 BCT",0

	; A bunch of debugging strings
	szBranch db "Branch instruction",13,10,0
	szTrue db "true",13,10,0
	szBuffer db 80 dup (0)
	szFormat db "Read %08x at pc=%08x",13,10,0
	szFormat2 db "Instruction type %d at %08x",13,10,0
	szFormat3 db "Writing %04x at %08x",13,10,0
	szFormat4 db "Branching with link from %08x to %08x",13,10,0
	szFormat5 db "mov pc,lr (%08x)",13,10,0
	szFormat6 db "stmfd %08x",13,10,0
	szFormat7 db "ldmfd %08x",13,10,0
	szFormat8 db "ldr r%d,=%d",13,10,0
	szFormat9 db "MUL r%d,%08x,%08x",13,10,0
	szFormat10 db "%08x",13,10,0
	szFormat11 db "%08x, %08x, %08x, %08x, %08x, PC=%08x",13,10,0
	szFormat12 db "Creating cache entry for address %08x",13,10,0
	szFormat13 db "Block has not yet been translated",13,10,0
	szFormat14 db "Starting execution of block at %08x",13,10,0
	szFormat15 db "Ending block with PC=%08x",13,10,0
	
	
	
.data?
	; General registers
	r0	dd ?	; r0
	r1	dd ?	; at
	r2	dd ?	; v0
	r3	dd ?	; v1
	r4	dd ?	; a0
	r5	dd ?	; a1
	r6	dd ?	; a2
	r7	dd ?	; a3
	r8	dd ?	; t0
	r9	dd ?	; t1
	r10	dd ?	; t2
	r11	dd ?	; t3
	r12	dd ?	; t4
	r13	dd ?	; t5	
	r14	dd ?	; t6	
	r15	dd ?	; t7	
	r16	dd ?	; s0
	r17	dd ?	; s1
	r18	dd ?	; s2
	r19	dd ?	; s3
	r20	dd ?	; s4
	r21	dd ?	; s5
	r22	dd ?	; s6
	r23	dd ?	; s7
	r24	dd ?	; t8
	r25	dd ?	; t9
	r26	dd ?	; k0
	r27	dd ?	; k1
	r28	dd ?	; gp
	r29	dd ?	; sp
	r30	dd ?	; s8	
	r31	dd ?	; ra	

	; Multiplication/division registers
	_hi	dd ?
	_lo	dd ?

	; Program counter, and other misc registers	
	PC	dd ?
	EPC	dd ?
	Cause	dd ?
	Status	dd ?

	UNKNOWN_INSTRUCTION EQU 0
	ARITHMETIC_OVERFLOW EQU 1

	; Coprocessor1 registers	
	f0	dd ?
	f1	dd ?
	f2	dd ?
	f3	dd ?
	f4	dd ?
	f5	dd ?
	f6	dd ?
	f7	dd ?
	f8	dd ?
	f9	dd ?
	f10	dd ?
	f11	dd ?
	f12	dd ?
	f13	dd ?
	f14	dd ?
	f15	dd ?
	f16	dd ?
	f17	dd ?
	f18	dd ?
	f19	dd ?
	f20	dd ?
	f21	dd ?
	f22	dd ?
	f23	dd ?
	f24	dd ?
	f25	dd ?
	f26	dd ?
	f27	dd ?
	f28	dd ?
	f29	dd ?
	f30	dd ?
	f31	dd ?

	; Coprocessor1 flag	
	fflag	dd ?
	
	cacheBlock dd ?
	pendingBranch dd ?
	branchOffset dd ?
	condBranch dd ?
	branchEdi dd ?
	regMap1 dd ?
	regMap2 dd ?
	hInst	dd ?
	hout	dd ?
	dummy 	dd ?
	lastPC 	dd ?
	lastCyc	dd ?
	physaddr dd ?
	nextInstr dd ?
	nextAddr dd ?
	zombie	_4ARGS ?
	mmu_read_byte _1ARGS ?
	mmu_read_word _1ARGS ?
	mmu_read_dword _1ARGS ?
	mmu_write_byte _2ARGS ?
	mmu_write_word _2ARGS ?
	mmu_write_dword _2ARGS ?
	mmu_get _1ARGS ?
	mmu_translate_address _1ARGS ?

	; Translation cache (64 entries)	
	cache cache_entry 64 DUP (<>)
	
.code


;########################################### Macros ##################################################

; Increase the cycle count (fixed point)
ADD_CYCLES MACRO num
	add cycle,num*200h
ENDM

	

COND_BRANCH MACRO jump
	LOCAL cond_true

	; Get Rs and Rt
	GET_ST

	; MOV EAX,Rs / CMP EAX,Rt
	MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
	OP_EAX_SR_PLUS_IMM32 03Bh,ecx,4,r0

	; Jump over the PC-update code if the condition was false	
	mov byte ptr [edi],jump
	add edi,2
	mov branchEdi,edi

	; Calculate the distance to jump	
	shl eax,2
	and eax,0FFFCh
	movsx eax,ax
	mov branchOffset,eax


	; Signal a pending branch to the translation function	
	mov pendingBranch,2
	mov condBranch,1
	
	ADD_CYCLES 2
	RETURN
ENDM



COND_BRANCH_ZR MACRO jump
	GET_S
	lea esi,[r0+ebx*4]
	mov word ptr [edi],3D83h	; 83 3D = cmp dword ptr x,imm8
	mov [edi+2],esi
	mov byte ptr [edi+6],0
	add edi,7
	
	mov byte ptr [edi],jump
	add edi,2
	mov branchEdi,edi
	
	shl eax,2
	and eax,0FFFCh
	movsx eax,ax
	mov branchOffset,eax
	
	mov pendingBranch,2
	mov condBranch,1
	
	ADD_CYCLES 2
	RETURN
ENDM


; Put Rd, Rs and Rt indexes in edx, ebx and ecx
GET_DST MACRO
	mov ebx,eax
	mov ecx,eax
	mov edx,eax
	shr ebx,21
	shr ecx,16
	shr edx,11
	and ebx,31
	and ecx,31
	and edx,31
ENDM

; Put Rd, Rs and Sa indexes in edx, ebx and ecx
GET_DSA MACRO
	mov ebx,eax
	mov ecx,eax
	mov edx,eax
	shr ebx,21
	shr ecx,6
	shr edx,11
	and ebx,31
	and ecx,31
	and edx,31
ENDM


; Put Rd, Rt and Sa indexes in edx, ebx and ecx
GET_DTA MACRO
	mov ebx,eax
	mov ecx,eax
	mov edx,eax
	shr ebx,16
	shr ecx,6
	shr edx,11
	and ebx,31
	and ecx,31
	and edx,31
ENDM


GET_ST MACRO
	mov ebx,eax
	mov ecx,eax
	shr ebx,21
	shr ecx,16
	and ebx,31
	and ecx,31
ENDM

GET_S MACRO
	mov ebx,eax
	shr ebx,21
	and ebx,31
ENDM

GET_T MACRO
	mov ecx,eax
	shr ecx,16
	and ecx,31
ENDM

GET_D MACRO
	mov edx,eax
	shr edx,11
	and edx,31
ENDM


; Put Rd, Rs and Rt indexes in edx, ebx and ecx
GET_COP_DST MACRO
	mov ebx,eax
	mov ecx,eax
	mov edx,eax
	shr ebx,11
	shr ecx,16
	shr edx,6
	and ebx,31
	and ecx,31
	and edx,31
ENDM

GET_COP_ST MACRO
	mov ebx,eax
	mov ecx,eax
	shr ebx,11
	shr ecx,16
	and ebx,31
	and ecx,31
ENDM


RETURN MACRO
	ret
ENDM




; MOV EAX,[reg*scale + imm]
MOV_EAX_SR_PLUS_IMM32 MACRO reg,scale,imm
	mov byte ptr [edi],0A1h
	lea esi,[reg*scale + imm]
	mov dword ptr [edi+1],esi
	add edi,5
ENDM


; MOV [reg*scale + imm],AL
MOV_SR_PLUS_IMM32_AL MACRO reg,scale,imm
	mov byte ptr [edi],0A2h		; A2 = mov r/m8,al
	lea esi,[reg*scale + imm]
	mov [edi+1],esi
	add edi,5
ENDM


; MOV [reg*scale + imm],AX
MOV_SR_PLUS_IMM32_AX MACRO reg,scale,imm
	mov word ptr [edi],0A366h	; 66 A3 = mov r/m16,ax
	lea esi,[reg*scale + imm]
	mov [edi+2],esi
	add edi,6
ENDM


; MOV [reg*scale + imm],EAX
MOV_SR_PLUS_IMM32_EAX MACRO reg,scale,imm
	mov byte ptr [edi],0A3h
	lea esi,[reg*scale + imm]
	mov dword ptr [edi+1],esi
	add edi,5
ENDM


; MOV EAX,[imm]
MOV_EAX_M32 MACRO m32
	mov byte ptr [edi],0A1h
	mov dword ptr [edi+1],offset m32
	add edi,5
ENDM


; MOV [imm],EAX
MOV_M32_EAX MACRO m32
	mov byte ptr [edi],0A3h
	mov dword ptr [edi+1],offset m32
	add edi,5
ENDM


; OP EAX,[reg*scale + imm]  (OP can be eg. ADD, SUB, AND, OR  etc)
OP_EAX_SR_PLUS_IMM32 MACRO op,reg,scale,imm
	mov word ptr [edi],op + 0500h
	lea esi,[reg*scale+imm]
	mov dword ptr [edi+2],esi
	add edi,6
ENDM


; CDQ / IDIV DWORD PTR [reg*scale + imm] / MOV _hi,EDX / MOV _lo,EAX
IDIV_SR_PLUS_IMM32 MACRO reg,scale,imm
	mov dword ptr [edi],003DF799h		; 99 = CDQ, F7 3D = IDIV r/m32
	lea esi,[reg*scale+imm]
	mov [edi+3],esi
	mov word ptr [edi+7],1589h		; 89 15 = mov r/m32,edx
	mov dword ptr [edi+9],offset _hi
	mov byte ptr [edi+13],0A3h		; A3 = mov r/m32,eax
	mov dword ptr [edi+14],offset _lo
	add edi,18
ENDM


; IMUL DWORD PTR [reg*scale + imm] / MOV _hi,EDX / MOV _lo,EAX
IMUL_SR_PLUS_IMM32 MACRO reg,scale,imm
	lea esi,[reg*scale+imm]
	mov word ptr [edi],2DF7h		; F7 2D = imul dword ptr x
	mov [edi+2],esi
	mov word ptr [edi+6],1589h		; 89 15 = mov r/m32,edx
	mov dword ptr [edi+8],offset _hi
	mov byte ptr [edi+12],0A3h		; A3 = mov r/m32,eax
	mov dword ptr [edi+13],offset _lo
	add edi,17
ENDM


ALU_OP MACRO type1,type2
	GET_DST
	.if ebx == edx
		MOV_EAX_SR_PLUS_IMM32 ecx,4,r0
		OP_EAX_SR_PLUS_IMM32 type1,edx,4,r0
	.else
		MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
		OP_EAX_SR_PLUS_IMM32 type2,ecx,4,r0
		MOV_SR_PLUS_IMM32_EAX edx,4,r0
	.endif
	ADD_CYCLES 1
	RETURN
ENDM


; MOV DWORD PTR [reg*scale + imm],imm
MOV_SR_PLUS_IMM32_IMM32 MACRO reg,scale,imm1,imm2
	mov word ptr [edi],05C7h
	lea esi,[reg*scale+imm1]
	mov [edi+2],esi
	mov dword ptr [edi+6],imm2
	add edi,10
ENDM


ALUI_OP MACRO op,reg1,reg2,scale,imm
	mov word ptr [edi],1D8Bh		; mov ebx,r/m32
	lea esi,[reg1*scale+imm]
	mov [edi+2],esi
	mov word ptr [edi+6],81h + op*100h
	mov dword ptr [edi+8],eax
	mov word ptr [edi+12],1D89h
	lea esi,[reg2*scale+imm]
	mov [edi+14],esi
	add edi,18
ENDM


SHIFT_MEM MACRO reg,scale,imm, shamt, op
	lea esi,[reg*scale+imm]
	mov word ptr [edi],0C1h + op*100h
	mov [edi+2],esi
	mov byte ptr [edi+6],shamt
	add edi,7
ENDM

SHIFT_EAX MACRO shamt,op
	mov word ptr [edi],0C1h + op*100h
	mov byte ptr [edi+2],shamt
	add edi,3
ENDM

OP_SHL EQU 0E0h
OP_SHR EQU 0E8h
OP_SAR EQU 0F8h


;######################################################################################################


LibMain PROC hInstDLL:DWORD, reason:DWORD, unused:DWORD

        .if reason == 1		;DLL_PROCESS_ATTACH
        	mov eax,hInstDLL
        	mov hInst,eax
		mov eax,TRUE
		ret

        .elseif reason == 0	;DLL_PROCESS_DETACH
        .elseif reason == 2	;DLL_THREAD_ATTACH
        .elseif reason == 3	;DLL_THREAD_DETACH
        .endif

        mov eax,FALSE
        ret
LibMain ENDP


inpb_null PROC
	ret
inpb_null ENDP
.data
inpb_handlers dd 256 DUP(OFFSET inpb_null)
.code

outpb_null PROC dat:DWORD
	ret
outpb_null ENDP
.data
outpb_handlers dd 256 DUP(OFFSET outpb_null)
.code	



;########################################### Opcodes ##################################################



op_and:
	ALU_OP 21h,23h


op_add:
	ALU_OP 01h,03h
	

op_addu:
	ALU_OP 01h,03h
	

op_break:


op_div:
op_divu:
	; Get Rs and Rt
	GET_ST
	; MOV EAX,Rs
	MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
	; IDIV Rt
	IDIV_SR_PLUS_IMM32 ecx,4,r0
	ADD_CYCLES 35
	RETURN

op_jalr:

op_jr:
	; Get Rs
	GET_S

	; MOV EAX,Rs		

	MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
	; MOV PC,EAX
	MOV_M32_EAX PC

	; Signal a pending unconditional branch. A branchOffset of 1 is used to specify
	; a register-indirect branch, as opposed to an absolute branch.
	mov branchOffset,1
	mov pendingBranch,2
	mov condBranch,0

	ADD_CYCLES 2
	RETURN
	
op_mfhi:
	; Get Rd
	GET_D
	; MOV EAX,_hi
	MOV_EAX_M32 _hi
	; MOV Rd,EAX
	MOV_SR_PLUS_IMM32_EAX edx,4,r0
	ADD_CYCLES 1
	RETURN

op_mflo:
	; Get Rd
	GET_D
	; MOV EAX,_lo
	MOV_EAX_M32 _lo
	; MOV Rd,EAX
	MOV_SR_PLUS_IMM32_EAX edx,4,r0
	ADD_CYCLES 1
	RETURN

op_mthi:
op_mtlo:

op_mult:
op_multu:
	GET_ST
	MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
	IMUL_SR_PLUS_IMM32 ecx,4,r0
	;mov eax,[r0+ebx*4]
	;imul dword ptr [r0+ecx*4]
	;mov _hi,edx
	;mov _lo,eax
	ADD_CYCLES 2
	RETURN
	
op_nor:
	GET_DST
	mov ebx,[r0+ebx*4]
	or ebx,[r0+ecx*4]
	not ebx

	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
op_or:
	ALU_OP 09h,0Bh


op_sll:
	GET_DTA
	.if ebx == edx
		SHIFT_MEM edx,4,r0, cl, 25h
	.else
		MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
		SHIFT_EAX cl,OP_SHL
		MOV_SR_PLUS_IMM32_EAX edx,4,r0
	.endif
	ADD_CYCLES 1
	RETURN

op_sllv:
	GET_DST
	;mov ebx,[r0+ebx*4]
	;mov ecx,[r0+ecx*4]
	;shl ebx,cl
	;mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_slt:
	GET_DST
	mov eax,[r0+ebx*4]
	cmp eax,[r0+ecx*4]
	setl bl
	and ebx,1
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
	
op_sltu:
	GET_DST
	mov eax,[r0+ebx*4]
	cmp eax,[r0+ecx*4]
	setb bl
	and ebx,1
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_sra:
	GET_DTA
	.if ebx == edx
		SHIFT_MEM edx,4,r0, cl, 3Dh
	.else
		MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
		SHIFT_EAX cl,OP_SAR
		MOV_SR_PLUS_IMM32_EAX edx,4,r0
	.endif
	ADD_CYCLES 1
	RETURN
	
op_srav:
	GET_DST
	mov ebx,[r0+ebx*4]
	mov ecx,[r0+ecx*4]
	sar ebx,cl
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
	
op_srl:
	GET_DTA
	.if ebx == edx
		SHIFT_MEM edx,4,r0, cl, 2Dh
	.else
		MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
		SHIFT_EAX cl,OP_SHR
		MOV_SR_PLUS_IMM32_EAX edx,4,r0
	.endif
	ADD_CYCLES 1
	RETURN
	
op_srlv:
	GET_DST
	mov ebx,[r0+ebx*4]
	mov ecx,[r0+ecx*4]
	shr ebx,cl
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
	
op_sub:
op_subu:
	ALU_OP 29h,2Bh
	
op_syscall:
	.if r2 == 1
		pusha
		invoke wsprintf,ADDR szBuffer,ADDR szFormat10,r4
		invoke lstrlen,ADDR szBuffer
		invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
		popa
	.endif
	ADD_CYCLES 1
	RETURN
	
op_xor:
	ALU_OP 31h,33h

op_unknown:
;	mov eax,PC
;	sub eax,4
;	mov EPC,eax
;	mov Cause,UNKNOWN_INSTRUCTION
	ADD_CYCLES 1
	RETURN



rtype_op:
	dd OFFSET op_sll
	dd OFFSET op_unknown	
	dd OFFSET op_srl
	dd OFFSET op_sra
	dd OFFSET op_sllv
	dd OFFSET op_unknown
	dd OFFSET op_srlv
	dd OFFSET op_srav
	dd OFFSET op_jr
	dd OFFSET op_jalr
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_syscall
	dd OFFSET op_break
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	
	dd OFFSET op_mfhi
	dd OFFSET op_mthi
	dd OFFSET op_mflo
	dd OFFSET op_mtlo
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_mult
	dd OFFSET op_multu
	dd OFFSET op_div
	dd OFFSET op_divu
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	

	dd OFFSET op_add
	dd OFFSET op_addu
	dd OFFSET op_sub
	dd OFFSET op_subu
	dd OFFSET op_and
	dd OFFSET op_or
	dd OFFSET op_xor
	dd OFFSET op_nor
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_slt
	dd OFFSET op_sltu
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	

	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	;dd OFFSET @@next_instr
	


opcode_00:

opcode_01:
	test eax,10000h
	jnz @@bgez
		COND_BRANCH_ZR 7Dh ;jl
	@@bgez:
		COND_BRANCH_ZR 7Ch ;jge
	
; J
opcode_02:
	; Caluclate branch offset
	and eax,003FFFFFFh
	shl eax,2
	mov branchOffset,eax

	; Signal a pending unconditional branch
	mov pendingBranch,2
	mov condBranch,0

	ADD_CYCLES 2
	RETURN

; JAL
opcode_03:
	; Calculate branch offset
	and eax,003FFFFFFh
	shl eax,2
	mov branchOffset,eax

	; Set R31 to the current PC	
	mov eax,PC
	mov word ptr [edi],05C7h
	mov dword ptr [edi+2],offset r31
	mov [edi+6],eax
	add edi,10

	; Signal a pending unconditional branch
	mov pendingBranch,2
	mov condBranch,0

	ADD_CYCLES 2
	RETURN

	
; BEQ rs,rt,imm16
opcode_04:
	COND_BRANCH 75h 	; 75 = jne

; BNE rs,rt,imm16
opcode_05:

	COND_BRANCH 74h 	; 74 = je

; BLEZ rs,imm16
opcode_06:
	COND_BRANCH_ZR 7Fh	; 7F = jg

; BGTZ rs,imm16	
opcode_07:
	COND_BRANCH_ZR 7Eh	; 7E = jle

; ADDI rt,rs,imm16
opcode_08:
;	GET_ST
;	and eax,0FFFFh
;	mov ebx,[r0+ebx*4]
;	add ebx,eax
;	mov [r0+ecx*4],ebx
;	ADD_CYCLES 1
;	RETURN

; ADDIU rt,rs,imm16
opcode_09:
	GET_ST
	and eax,0FFFFh
	ALUI_OP 0C3h,ebx,ecx,4,r0
	ADD_CYCLES 1
	RETURN

; SLTI rt,rs,imm16	
opcode_0A:
	GET_ST
	and eax,0FFFFh
	cmp eax,[r0+ebx*4]
	setl bl
	and ebx,1
	mov [r0+ecx*4],ebx
	ADD_CYCLES 1
	RETURN

; SLTIU rt,rs,imm16
opcode_0B:
	GET_ST
	and eax,0FFFFh
	cmp eax,[r0+ebx*4]
	setb bl
	and ebx,1
	mov [r0+ecx*4],ebx
	ADD_CYCLES 1
	RETURN

; ANDI rt,rs,imm16
opcode_0C:
	GET_ST
	and eax,0FFFFh
	ALUI_OP 0E3h,ebx,ecx,4,r0
	ADD_CYCLES 1
	RETURN



; ORI rt,rs,imm16
opcode_0D:
	GET_ST
	and eax,0FFFFh
	.if (eax == 0 && ebx == 0)
		MOV_SR_PLUS_IMM32_IMM32 ecx,4,r0,0
	.else
		ALUI_OP 0CBh,ebx,ecx,4,r0
	.endif
	ADD_CYCLES 1
	RETURN

; XORI rt,rs,imm16	
opcode_0E:
	GET_ST
	and eax,0FFFFh
	ALUI_OP 0F3h,ebx,ecx,4,r0
	ADD_CYCLES 1
	RETURN
	
; LUI rt,imm16
opcode_0F:
	GET_T
	and eax,0FFFFh
	

	mov word ptr [edi],2581h	; and dword ptr x,imm32
	lea esi,[r0+ecx*4]
	mov [edi+2],esi
	mov dword ptr [edi+6],0FFFFh
	add edi,10
	
	shl eax,16
	mov word ptr [edi],0D81h	; or dword ptr x,imm32
	mov [edi+2],esi
	mov dword ptr [edi+6],eax
	add edi,10

	ADD_CYCLES 1
	RETURN

	
opcode_10:

; Coprocessor 1 instruction
opcode_11:
	mov ebx,eax
	and ebx,03E00000h
	.if ebx == 02000000h	; Single precision
		GET_COP_DST
		and eax,03Fh
		; ADD.S rd,rs,rt
		.if eax == 000000h
			fld dword ptr [f0+ebx*4]
			fadd dword ptr [f0+ecx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 2
		; SUB.S rd,rs,rt
		.elseif eax == 000001h
			fld dword ptr [f0+ebx*4]
			fsub dword ptr [f0+ecx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 2
		; MUL.S rd,rs,rt
		.elseif eax == 000002h
			fld dword ptr [f0+ebx*4]
			fmul dword ptr [f0+ecx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 4
		; DIV.S rd,rs,rt
		.elseif eax == 000003h
			fld dword ptr [f0+ebx*4]
			fdiv dword ptr [f0+ecx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 40
		; ABS.S rd,rs
		.elseif eax == 000005h
			fld dword ptr [f0+ebx*4]
			fabs
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 2
		; NEG.S rd,rs
		.elseif eax == 000007h
			fld dword ptr [f0+ebx*4]
			fchs
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 2
		; CVT.W.S rd,rs
		.elseif eax == 000024h
			fld dword ptr [f0+ebx*4]
			fistp dword ptr [f0+edx*4]
			ADD_CYCLES 1
		.endif
	.elseif ebx == 0
		GET_COP_ST
		and eax,03Fh
		; MFC1 rt,rs
		.if eax == 000000h
			mov eax,[f0+ebx*4]
			mov [r0+ecx*4],eax
			ADD_CYCLES 1
		.endif
	.elseif ebx == 800000h
		GET_COP_ST
		and eax,03Fh
		; MTC1 rt,rs
		.if eax == 0
			mov eax,[r0+ebx*4]
			mov [f0+ecx*4],eax
			ADD_CYCLES 1
		.endif
	.elseif ebx == 2800000h
		GET_COP_DST
		and eax,03Fh
		; CVT.S.W rd,rs,rt
		.if eax == 20h
			fild dword ptr [f0+ebx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 1
		.endif
	.endif
	RETURN
	
opcode_12:
opcode_13:
opcode_14:
opcode_15:
opcode_16:
opcode_17:
opcode_18:
opcode_19:
opcode_1A:
opcode_1B:
opcode_1C:
opcode_1D:
opcode_1E:
opcode_1F:


LOAD_OP MACRO func
	GET_ST
	and eax,0FFFFh
	
	MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
	mov byte ptr [edi],05h
	mov [edi+1],eax

	mov byte ptr [edi+5],50h	; push eax
	add edi,6
	
	mov word ptr [edi],1D8Bh
	mov dword ptr [edi+2],offset func
	mov word ptr [edi+6],0D3FFh
	add edi,8
ENDM

; LB rt,imm(rs)
opcode_20:
	LOAD_OP mmu_read_byte
	mov word ptr [edi],0BE0Fh
	mov byte ptr [edi+2],0C0h
	add edi,3
	MOV_SR_PLUS_IMM32_EAX ecx,4,r0
	ADD_CYCLES 2
	RETURN

; old code
	GET_ST
	and eax,0FFFFh
	add eax,[r0+ebx*4]
	push ecx
	push ebp
	invoke mmu_read_byte,eax
	pop ebp
	pop ecx
	movsx eax,al
	mov [r0+ecx*4],eax
	ADD_CYCLES 2
	RETURN

; LH rt,imm(rs)	
opcode_21:
	GET_ST
	and eax,0FFFFh
	add eax,[r0+ebx*4]
	push ecx
	push ebp
	invoke mmu_read_word,eax
	pop ebp
	pop ecx
	movsx eax,ax
	mov [r0+ecx*4],eax
	ADD_CYCLES 2
	RETURN
	
opcode_22:

; LW rt,imm(rs)
opcode_23:
	LOAD_OP mmu_read_dword
	MOV_SR_PLUS_IMM32_EAX ecx,4,r0
	ADD_CYCLES 2
	RETURN
	
; LBU rt,imm(rs)
opcode_24:
	LOAD_OP mmu_read_byte
	MOV_SR_PLUS_IMM32_AL ecx,4,r0
	ADD_CYCLES 2
	RETURN


; LHU rt,imm(rs)	
opcode_25:
	LOAD_OP mmu_read_word
	MOV_SR_PLUS_IMM32_AX ecx,4,r0
	ADD_CYCLES 2
	RETURN
	
opcode_26:
opcode_27:

STORE_OP MACRO func
	GET_ST
	and eax,0FFFFh
	
	MOV_EAX_SR_PLUS_IMM32 ebx,4,r0
	mov byte ptr [edi],05h
	mov [edi+1],eax
	add edi,5

	mov word ptr [edi],35FFh	; push dword ptr x
	lea esi,[r0+ecx*4]
	mov [edi+2],esi
	mov byte ptr [edi+6],50h	; push eax
	add edi,7
	
	mov word ptr [edi],1D8Bh
	mov dword ptr [edi+2],offset func
	mov word ptr [edi+6],0D3FFh
	add edi,8
	
	ADD_CYCLES 2
	RETURN
ENDM


; SB rt,imm(rs)
opcode_28:
	STORE_OP mmu_write_byte

; SH rt,imm(rs)	
opcode_29:
	STORE_OP mmu_write_word


opcode_2A:

; SW rt,imm(rs)
opcode_2B:
	STORE_OP mmu_write_dword

opcode_2C:
opcode_2D:
opcode_2E:
opcode_2F:

opcode_30:

; L.S
opcode_31:
	GET_ST
	and eax,0FFFFh
	add eax,[r0+ebx*4]
	push ecx
	push ebp
	invoke mmu_read_dword,eax
	pop ebp
	pop ecx
	mov [f0+ecx*4],eax
	ADD_CYCLES 2
	RETURN	

opcode_32:
opcode_33:
opcode_34:
opcode_35:
opcode_36:
opcode_37:
opcode_38:
opcode_39:
opcode_3A:
opcode_3B:
opcode_3C:
opcode_3D:
opcode_3E:
opcode_3F:

	
opcode:
	dd OFFSET opcode_00
	dd OFFSET opcode_01
	dd OFFSET opcode_02
	dd OFFSET opcode_03
	dd OFFSET opcode_04
	dd OFFSET opcode_05
	dd OFFSET opcode_06
	dd OFFSET opcode_07
	dd OFFSET opcode_08
	dd OFFSET opcode_09
	dd OFFSET opcode_0A
	dd OFFSET opcode_0B
	dd OFFSET opcode_0C
	dd OFFSET opcode_0D
	dd OFFSET opcode_0E
	dd OFFSET opcode_0F
	dd OFFSET opcode_10
	dd OFFSET opcode_11
	dd OFFSET opcode_12
	dd OFFSET opcode_13
	dd OFFSET opcode_14
	dd OFFSET opcode_15
	dd OFFSET opcode_16
	dd OFFSET opcode_17
	dd OFFSET opcode_18
	dd OFFSET opcode_19
	dd OFFSET opcode_1A
	dd OFFSET opcode_1B
	dd OFFSET opcode_1C
	dd OFFSET opcode_1D
	dd OFFSET opcode_1E
	dd OFFSET opcode_1F
	dd OFFSET opcode_20
	dd OFFSET opcode_21
	dd OFFSET opcode_22
	dd OFFSET opcode_23
	dd OFFSET opcode_24
	dd OFFSET opcode_25
	dd OFFSET opcode_26
	dd OFFSET opcode_27
	dd OFFSET opcode_28
	dd OFFSET opcode_29
	dd OFFSET opcode_2A
	dd OFFSET opcode_2B
	dd OFFSET opcode_2C
	dd OFFSET opcode_2D
	dd OFFSET opcode_2E
	dd OFFSET opcode_2F
	dd OFFSET opcode_30
	dd OFFSET opcode_31
	dd OFFSET opcode_32
	dd OFFSET opcode_33
	dd OFFSET opcode_34
	dd OFFSET opcode_35
	dd OFFSET opcode_36
	dd OFFSET opcode_37
	dd OFFSET opcode_38
	dd OFFSET opcode_39
	dd OFFSET opcode_3A
	dd OFFSET opcode_3B
	dd OFFSET opcode_3C
	dd OFFSET opcode_3D
	dd OFFSET opcode_3E
	dd OFFSET opcode_3F
	

;######################################################################################################


cpu_init PROC commlink:DWORD	
	mov eax,commlink
	mov zombie,eax

	invoke GetStdHandle,STD_OUTPUT_HANDLE
	mov hout,eax

	pusha	
	mov edi,offset r0
	mov ecx,32
	mov eax,0
	rep stosd
	popa
	
	ret
cpu_init ENDP



free_cache PROC
	pusha
	
	mov ecx,0
	@@free_blocks:
		cmp (cache_entry PTR [cache + ecx]).inUse,0
		je @@line_free
			push ecx
			mov ebx,(cache_entry PTR [cache + ecx]).codePtr
			invoke LocalFree,ebx
			pop ecx
		@@line_free:
		mov (cache_entry PTR [cache + ecx]).inUse,0
		mov (cache_entry PTR [cache + ecx]).address,0
		add ecx,sizeof(cache_entry)
		cmp ecx,64*sizeof(cache_entry)
	jne @@free_blocks

	popa
	ret
free_cache ENDP



cpu_reset PROC
	pusha

	invoke zombie,ASK,MMU_READ_BYTE_CALLBACK,0,0
	mov mmu_read_byte,eax
	invoke zombie,ASK,MMU_READ_WORD_CALLBACK,0,0
	mov mmu_read_word,eax
	invoke zombie,ASK,MMU_READ_DWORD_CALLBACK,0,0
	mov mmu_read_dword,eax

	invoke zombie,ASK,MMU_WRITE_BYTE_CALLBACK,0,0
	mov mmu_write_byte,eax
	invoke zombie,ASK,MMU_WRITE_WORD_CALLBACK,0,0
	mov mmu_write_word,eax
	invoke zombie,ASK,MMU_WRITE_DWORD_CALLBACK,0,0
	mov mmu_write_dword,eax

	invoke zombie,ASK,MMU_TRANSLATE_ADDRESS_CALLBACK,0,0
	mov mmu_translate_address,eax

	; Clear registers
	mov edi,offset r0
	mov ecx,32
	mov eax,0
	rep stosd

	; Clear translation cache
	call free_cache
	
	popa
	ret
cpu_reset ENDP


cpu_close PROC
	call free_cache
	ret
cpu_close ENDP


cpu_config PROC hWnd:DWORD
	ret
cpu_config ENDP


cpu_set_io_handler PROC mode:DWORD,port:DWORD,pfhandler:DWORD
	push ebx

	mov eax,port
	mov ebx,pfhandler
	
	.if mode==INPUT	
		mov [inpb_handlers + eax*4],ebx
	.else
		mov [outpb_handlers + eax*4],ebx
	.endif
	
	pop ebx
	ret
cpu_set_io_handler ENDP



cpu_get PROC what:DWORD
	.if what == CPU_CYCLE_COUNT
		mov eax,cycle
	.elseif what == CPU_CYCLE_SCALING
		mov eax,200h
	.elseif what == CPU_SET_IO_CALLBACK
		mov eax,offset cpu_set_io_handler
	.elseif what == CPU_IRQ_CALLBACK
		mov eax,offset cpu_irq
	.elseif what == CPU_INPB_CALLBACK
		mov eax,offset cpu_inpb
	.elseif what == CPU_OUTPB_CALLBACK
		mov eax,offset cpu_outpb
	.elseif what == CPU_SPEED
		mov eax,speed
	.elseif what == CPU_REG_PC
		mov eax,PC
	.elseif what == CPU_REG
		pusha
		invoke wsprintf,ADDR szBuffer,ADDR szFormat11,r0,r2,r3,r4,r5,PC
		invoke lstrlen,ADDR szBuffer
		invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
		popa
		
	;	mov eax,param1
	;	and eax,0Fh
	;	mov eax,[r0+eax*4]
	.elseif what == PLUGIN_NAME
		mov eax,offset cpuName
	.endif
	ret
cpu_get ENDP


cpu_set PROC what:DWORD,param1:DWORD,param2:DWORD
	.if what == CPU_REG_PC
		mov eax,param1
		mov PC,eax
	.elseif what == CPU_REG_SP
		mov eax,param1
		mov r29,eax
	.elseif what == CPU_CYCLE_COUNT
		mov eax,param1
		mov cycle,eax
	.elseif what == CPU_BREAKPOINT
		mov eax,param1
		mov breakpoint,eax
	.elseif what == CPU_SPEED
		mov eax,param1
		mov speed,eax
	.elseif what == CPU_IO_HANDLER
		mov eax,param1
		and param1,0FFh
		shr eax,8
		invoke cpu_set_io_handler,eax,param1,param2
	.endif
	ret
cpu_set ENDP



cpu_outpb PROC port:DWORD,dat:DWORD
	push ecx
	push dword ptr dat
	mov eax,port
	call dword ptr [outpb_handlers + eax*4]
	pop ecx
	ret
cpu_outpb ENDP


cpu_inpb PROC port:DWORD
	push ecx
	mov eax,port
	call dword ptr [inpb_handlers + eax*4]
	pop ecx
	ret
cpu_inpb ENDP


cpu_irq PROC address:DWORD
	ret
cpu_irq ENDP


cpu_execute PROC
	ret
cpu_execute ENDP



MOV_SR_PLUS_IMM32_R32 MACRO reg1,scale,imm,reg2
	lea esi,[reg1*scale+imm]
	mov byte ptr [edi],89h
	mov byte ptr [edi+1],reg2*8 + 5
	mov [edi+2],esi
	add edi,6
ENDM



BRANCH MACRO
	.if condBranch == 0
		; A branchOffset or 1 (one) is used to signal a register-indirect jump (JR) 
		.if branchOffset != 1
			mov eax,PC
			sub eax,4
			and eax,0F0000000h
			or eax,branchOffset
	
			mov word ptr [edi],05C7h		; mov r/m32,imm32
			mov dword ptr [edi+2],offset PC
			mov [edi+6],eax
			add edi,10
		.endif
	.else

		
		mov eax,PC
		sub eax,4
		add eax,branchOffset
		mov word ptr [edi],05C7h
		mov dword ptr [edi+2],offset PC
		mov [edi+6],eax
		mov word ptr [edi+10],0AEBh	; jmp +10
		add edi,12

		mov eax,branchEdi
		mov ebx,edi
		sub ebx,eax
		mov [eax-1],bl			; set jump distance

		mov word ptr [edi],05C7h
		mov eax,PC
		mov dword ptr [edi+2],offset PC
		mov [edi+6],eax
		add edi,10
	.endif

	;.if regMap1 != -1
	;	mov eax,regMap1
	;	MOV_SR_PLUS_IMM32_R32 eax,4,r0,7
	;.endif
	
	; mov ebx,offset _block_done / jmp ebx
	mov byte ptr [edi],0BBh
	mov dword ptr [edi+1],offset _block_done
	mov word ptr [edi+5],0E3FFh
	add edi,7
ENDM



;write_block_info PROC
;	pusha
;	invoke wsprintf,ADDR szBuffer,ADDR szFormat14,0
;	invoke lstrlen,ADDR szBuffer
;	invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
;	popa
;write_block_info ENDP



cache_lookup PROC address:DWORD

	mov eax,address

	; Use bits 2..7 as an index
	and eax,0FFh
	shr eax,2

	; Multiply by 16 (the size of each cache entry)	
	shl eax,4
	mov cacheBlock,eax
	
	cmp (cache_entry PTR [cache + eax]).inUse,0
	jnz @@line_used
		; Allocate space for 2K words
		invoke LocalAlloc,LPTR,8192

		mov ebx,cacheBlock
		mov (cache_entry PTR [cache + ebx]).codePtr,eax
		mov (cache_entry PTR [cache + ebx]).inUse,1
	@@line_used:

	; Check if this block has already been translated
	mov ebx,cacheBlock
	mov ecx,address
	cmp ecx,(cache_entry PTR [cache + ebx]).address
	jz @@already_cached

	mov (cache_entry PTR [cache + ebx]).address,ecx
	push dword ptr cycle
	mov cycle,0
	mov pendingBranch,0
	mov edi,(cache_entry PTR [cache + ebx]).codePtr
	mov eax,address
	mov PC,eax
	mov regMap1,-1
	
	@@translate:
		push edi
		
		; Read one instruction from memory
		invoke mmu_read_dword,PC
		
		pop edi
		add PC,4
	
		test eax,0FC000000h
		jnz @@xlate_not_rtype
			; R-type instruction
			mov ebx,eax
			and ebx,03Fh
			call dword ptr [rtype_op + ebx*4]
			jmp @@xlate_next_instr
		@@xlate_not_rtype:
			; I- or J-type instruction
			mov ebx,eax
			shr ebx,26
			and ebx,03Fh
			call dword ptr [opcode + ebx*4]
		@@xlate_next_instr:
		
		; Check for pending branches
		.if pendingBranch > 0
			dec dword ptr [pendingBranch]
			jnz @@not_zero
				BRANCH
				jmp @@translation_done
			@@not_zero:
		.endif
	
		jmp @@translate
	
	@@translation_done:
	mov ecx,cycle
	mov ebx,cacheBlock
	mov (cache_entry PTR [cache + ebx]).cycles,ecx
	mov eax,(cache_entry PTR [cache + ebx]).address
	mov PC,eax
	pop dword ptr cycle
		
	@@already_cached:	

	mov eax,cacheBlock	
	ret
cache_lookup ENDP



PUBLIC cpu_execute_until
cpu_execute_until: ; PROC threshold:DWORD
	pusha
	
	mov ecx,cycle

	@@exe_loop:
		cmp ecx,[esp+36]	
		jge @@exe_done		; Is the cycle count larger than the threshold ?

		invoke cache_lookup,PC
		mov cacheBlock,eax

		mov ebx,(cache_entry PTR [cache + eax]).codePtr
		jmp ebx 
		
		_block_done::
		
		mov eax,cacheBlock
		mov ecx,cycle
		add ecx,(cache_entry PTR [cache + eax]).cycles
		mov cycle,ecx

		jmp @@exe_loop

	@@exe_done:
	popa
	mov eax,cycle
	ret 4


	

END LibMain
