@ crt0 /Mic, 2004
@ Copied from Jeff Frohwein's crt0, with some modifications


.TEXT
.GLOBAL _start

	
.equ Cpu_Mode_USR,0x10
.equ Cpu_Mode_IRQ,0x12
.equ Cpu_Mode_SVC,0x13

.equ Cpu_Stack_USR,0x03007F00	@ USR stack adress
.equ Cpu_Stack_IRQ,0x03007FA0	@ IRQ stack adress
.equ Cpu_Stack_SVC,0x03007FE0	@ SVC stack adress
	

_start:

@ Clear IWRAM
        mov     r0,#3
        mov     r0,r0,lsl #24         			@ r0 = 0x3000000
        ldr     r1,=__sp_usr_offset - 16
        bl      ClearMem

@ Copy initialized data (data section) from LMA to VMA (ROM to RAM)
        ldr     r1,=__data_lma
        ldr     r2,=__data_start
        ldr     r4,=__data_end
        bl      CopyMemChk

@ Copy internal work ram (iwram section) from LMA to VMA (ROM to RAM)
        ldr     r1,=__iwram_lma
        ldr     r2,=__iwram_start
        ldr     r4,=__iwram_end
        bl      CopyMemChk

@ Copy internal work ram overlay 0 (iwram0 section) from LMA to VMA (ROM to RAM)
        ldr     r2,=__load_stop_iwram0
        ldr     r1,=__load_start_iwram0
        subs    r3,r2,r1         @ Is there any data to copy?
        beq     CIW0Skip         @ no

        ldr     r2,=__iwram_overlay_start
        bl      CopyMem
CIW0Skip:

@ Copy external work ram (ewram section) from LMA to VMA (ROM to RAM)
        ldr     r1,=__ewram_lma
        ldr     r2,=__ewram_start
        ldr     r4,=__ewram_end
        bl      CopyMemChk

@ Copy external work ram overlay 0 (ewram0 section) from LMA to VMA (ROM to RAM)
        ldr     r2,=__load_stop_ewram0
        ldr     r1,=__load_start_ewram0
        subs    r3,r2,r1         @ Is there any data to copy?
        beq     CEW0Skip         @ no

        ldr     r2,=__ewram_overlay_start
        bl      CopyMem
CEW0Skip:



	ldr 	sp,=Cpu_Stack_USR		@ Setup USR stack pointer
	
	b	XgsMain				@ Now branch to start of C code
	
final_return_loop:
	b 	final_return_loop		@ In case of a return...




@ Clear memory to 0x00 if length != 0
@ r0 = Start Address
@ r1 = Length

ClearMem:
        mov     r2,#3            @ These commands are used in cases where
        add     r1,r1,r2         @ the length is not a multiple of 4,
        bics    r1,r1,r2         @ even though it should be.

        beq     ClearMX          @ Length is zero so exit

        mov     r2,#0
ClrLoop:
        str     r2,[r0],#4
        subs    r1,r1,#4
        bne     ClrLoop
ClearMX:
        mov	pc,lr
        
        
        

@ Copy memory if length != 0
@ r1 = Source Address
@ r2 = Dest Address
@ r4 = Dest Address + Length

CopyMemChk:
        sub     r3,r4,r2         @ Is there any data to copy?

@ Copy memory
@ r1 = Source Address
@ r2 = Dest Address
@ r3 = Length

CopyMem:
        mov     r0,#3            @ These commands are used in cases where
        add     r3,r3,r0         @ the length is not a multiple of 4,
        bics    r3,r3,r0         @ even though it should be.

        beq     CIDExit          @ Length is zero so exit

CIDLoop:
        ldr     r0,[r1],#4
        str     r0,[r2],#4
        subs    r3,r3,#4
        bne     CIDLoop
CIDExit:
mov	pc,lr



.ALIGN
.POOL
.END
 	